



def macro_def(mad):


    mad.input('''ptc_twiss_macro(order, dp, slice_flag): macro = {
      ptc_create_universe;
      ptc_create_layout, time=false, model=2, exact=true, method=6, nst=3;
      IF (slice_flag == 1){
        select, flag=ptc_twiss, clear;
        select, flag=ptc_twiss, column=name,keyword,s,l,x,px,beta11,beta22,disp1,k1l;
        ptc_twiss, closed_orbit, icase=56, no=order, deltap=dp, table=ptc_twiss, summary_table=ptc_twiss_summary, slice_magnets=true;
      }
      ELSE{
        select, flag=ptc_twiss, clear;
        select, flag=ptc_twiss, column=name,keyword,s,x,px,beta11,alfa11,beta22,alfa22,disp1,disp2,mu1,mu2,energy,l,angle,K1L,K2L,K3L,HKICK,SLOT_ID;    
        ptc_twiss, closed_orbit, icase=56, no=order, deltap=dp, table=ptc_twiss, summary_table=ptc_twiss_summary, normal;
      }
      ptc_end;
    };

    ''')

    mad.input(''' 
    write_ptc_twiss(filename) : macro = {
        write, table=ptc_twiss, file=filename;
    };
    ''')

    mad.input('''match_tunes(qx, qy, dqx, dqy) : macro = {

      match, sequence=ps;
              vary, name = kf, step=0.00001;
              vary, name = kd, step=0.00001;
              constraint, range = #E, mux = qx + dqx, muy = qy + dqy;
      jacobian,calls=50000,bisec=3, tolerance = 1E-20;
      ENDMATCH;

    };

    ''')

    mad.input('''tunes_leq_knob(): macro = {

      kf := kf_0 + dkf_x * qx_leq + dkf_y * qy_leq;
      kd := kd_0 + dkd_x * qx_leq + dkd_y * qy_leq;

    };

    ''')

    mad.input('''tunes_leq_knob_factors(qx0, qy0): macro = {

      kf_0 = kf; 
      kd_0 = kd;

      dq = 1E-2;

      qx0 = qx0 ; ! lef out a +6
      qy0 = qy0 ; ! lef out a +6 because in the code  it is actually already there

      exec, match_tunes(qx0, qy0, 0, 0);

      kf_1 = kf; 
      kd_1 = kd;

      exec, match_tunes(qx0, qy0, dq, 0);

      dkf_x = (kf - kf_1) / dq;
      dkd_x = (kd - kd_1) / dq;

      exec, match_tunes(qx0, qy0, 0, dq);

      dkf_y = (kf - kf_1) / dq;
      dkd_y = (kd - kd_1) / dq;  

      qx_leq = 0;
      qy_leq = 0;

      exec, tunes_leq_knob;

    };
    ''')

    mad.input('''write_str_file(filename): macro = {
      assign, echo = filename;
      print, text = "/**********************************************************************************";
      print, text = "*                             SBENDs and MULTIPOLES in MUs";
      print, text = "***********************************************************************************/";
      print, text = "";
      value, k1PRBHF, k1PRBHD, k2PRBHF, k2PRBHD, k2PRMP, k2PRMPJ, k3PRMPF, k3PRMPD;
      print, text = "";
      print, text = "/**********************************************************************************";
      print, text = "*                                    PFW and F8L";
      print, text = "***********************************************************************************/";
      print, text = "";
      value, k1PRPFWF, k1PRPFWD, k2PRPFWF, k2PRPFWD, k3PRPFWF, k3PRPFWD, k1PRF8L;
      print, text = "";
      print, text = "/**********************************************************************************";
      print, text = "*                                    Injection dipoles";
      print, text = "***********************************************************************************/";
      print, text = "";
      value, kPIBSW26;
      value, kPIBSW40, kPIBSW41, kPIBSW42, kPIBSW43, kPIBSW44;
      print, text = "";
      print, text = "/**********************************************************************************";
      print, text = "*                                    Extraction dipoles";
      print, text = "***********************************************************************************/";
      print, text = "";
      value, kPEBSW12, kPEBSW14, kPEBSW20, kPEBSW22;
      value, kPEBSW23, kPEBSW57;
      print, text = ""; 
      print, text = "/**********************************************************************************";
      print, text = "*                                      Quadrupoles";
      print, text = "***********************************************************************************/";
      print, text = "";
      value, kF, kD, kPRQSE, kPEQKE16, kPIQLB;
      print, text = "";
      print, text = "/**********************************************************************************";
      print, text = "*                                       Sextupoles";
      print, text = "***********************************************************************************/";
      print, text = "";
      value, kPRXNO39, kPRXNO55, kPRXNO, kPRXSE;
      print, text = "";
      print, text = "/**********************************************************************************";
      print, text = "*                                       Octupoles";
      print, text = "***********************************************************************************/";
      print, text = "";
      value, kPRONO39, kPRONO55, kPRODN;
      print, text = "";
      print, text = "/**********************************************************************************";
      print, text = "*                                   KNOBS";
      print, text = "***********************************************************************************/";
      IF (dkPIBSW40_x <> 0.) {
        print,  text = "";   
        print,  text = "!-------------------------------------------------";      
        print,  text = "! Proton injection bump"; 
        print,  text = "!-------------------------------------------------";
        print,  text = "";  
        print, text = "bsw42_x_mm_abs_active  =              1 ;";         
        printf, text = "bsw42_x_mm_abs     = %F ;", value = x0*1e3;
        print, text = "bsw42_x_mm         =       0.0000000000 ;";
        print, text = "bsw42_px_mrad      =       0.0000000000 ;";
        print,  text = "";      

        IF (second_injection == 1){
          printf, text = "bsw42.2_x_mm_abs   = %F ;", value = x0*1e3;
          print, text = "bsw42.2_x_mm       =       0.0000000000 ;";
          print, text = "bsw42.2_px_mrad    =       0.0000000000 ;";
          print,  text = "";  
        }

        printf, text = "dkPIBSW40_x_abs    = %F ;", value = dkPIBSW40_x_abs;
        printf, text = "dkPIBSW41_x_abs    = %F ;", value = dkPIBSW41_x_abs;
        printf, text = "dkPIBSW42_x_abs    = %F ;", value = dkPIBSW42_x_abs;
        printf, text = "dkPIBSW43_x_abs    = %F ;", value = dkPIBSW43_x_abs;
        printf, text = "dkPIBSW44_x_abs    = %F ;", value = dkPIBSW44_x_abs;
        print,  text = "";      
        printf, text = "dkPIBSW40_x        = %F ;", value = dkPIBSW40_x;
        printf, text = "dkPIBSW41_x        = %F ;", value = dkPIBSW41_x;
        printf, text = "dkPIBSW42_x        = %F ;", value = dkPIBSW42_x;
        printf, text = "dkPIBSW43_x        = %F ;", value = dkPIBSW43_x;
        printf, text = "dkPIBSW44_x        = %F ;", value = dkPIBSW44_x;
        print,  text = "";      
        printf, text = "dkPIBSW40_px       = %F ;", value = dkPIBSW40_px;
        printf, text = "dkPIBSW41_px       = %F ;", value = dkPIBSW41_px;
        printf, text = "dkPIBSW42_px       = %F ;", value = dkPIBSW42_px;
        printf, text = "dkPIBSW43_px       = %F ;", value = dkPIBSW43_px;
        printf, text = "dkPIBSW44_px       = %F ;", value = dkPIBSW44_px;
        print,  text = "";      
        print, text = "kPIBSW40 := dkPIBSW40_x_abs * bsw42_x_mm_abs + dkPIBSW40_x * bsw42_x_mm + dkPIBSW40_px * bsw42_px_mrad;";
        print, text = "kPIBSW41 := dkPIBSW41_x_abs * bsw42_x_mm_abs + dkPIBSW41_x * bsw42_x_mm + dkPIBSW41_px * bsw42_px_mrad;";
        print, text = "kPIBSW42 := dkPIBSW42_x_abs * bsw42_x_mm_abs + dkPIBSW42_x * bsw42_x_mm + dkPIBSW42_px * bsw42_px_mrad;";
        print, text = "kPIBSW43 := dkPIBSW43_x_abs * bsw42_x_mm_abs + dkPIBSW43_x * bsw42_x_mm + dkPIBSW43_px * bsw42_px_mrad;";
        print, text = "kPIBSW44 := dkPIBSW44_x_abs * bsw42_x_mm_abs + dkPIBSW44_x * bsw42_x_mm + dkPIBSW44_px * bsw42_px_mrad;";
        print,  text = "";  

        IF (second_injection == 1){

          print, text = "kPI2BSW40 := dkPIBSW40_x_abs * bsw42.2_x_mm_abs + dkPIBSW40_x * bsw42.2_x_mm + dkPIBSW40_px * bsw42.2_px_mrad;";
          print, text = "kPI2BSW41 := dkPIBSW41_x_abs * bsw42.2_x_mm_abs + dkPIBSW41_x * bsw42.2_x_mm + dkPIBSW41_px * bsw42.2_px_mrad;";
          print, text = "kPI2BSW42 := dkPIBSW42_x_abs * bsw42.2_x_mm_abs + dkPIBSW42_x * bsw42.2_x_mm + dkPIBSW42_px * bsw42.2_px_mrad;";
          print, text = "kPI2BSW43 := dkPIBSW43_x_abs * bsw42.2_x_mm_abs + dkPIBSW43_x * bsw42.2_x_mm + dkPIBSW43_px * bsw42.2_px_mrad;";
          print, text = "kPI2BSW44 := dkPIBSW44_x_abs * bsw42.2_x_mm_abs + dkPIBSW44_x * bsw42.2_x_mm + dkPIBSW44_px * bsw42.2_px_mrad;";
          print,  text = "";  

        }

      }
      ELSE{
        print,  text = "";   
        print,  text = "!-------------------------------------------------";      
        print,  text = "! Proton injection bump"; 
        print,  text = "!-------------------------------------------------";
        print,  text = "";   
        print, text = "bsw42_x_mm_abs_active  =              0 ;";
        print,  text = "";   
      }

      IF (dkPISMH26_x_abs <> 0.) {
        print,  text = "";   
        print,  text = "!-------------------------------------------------";      
        print,  text = "! Ion injection bump"; 
        print,  text = "!-------------------------------------------------";
        print,  text = "";           
        print, text = "smh26_x_mm_abs_active  =              1 ;";
        printf, text = "smh26_x_mm_abs      = %F ;", value = x0*1e3;
        print,  text = "";      
        printf, text = "dkPISMH26_x_abs     = %F ;", value = dkPISMH26_x_abs;
        print,  text = "";      
        print, text = "kPIBSW26 := dkPISMH26_x_abs * smh26_x_mm_abs;";
        print,  text = "";  
      } 
      ELSE{
        print,  text = "";   
        print,  text = "!-------------------------------------------------";      
        print,  text = "! Ion injection bump"; 
        print,  text = "!-------------------------------------------------";
        print,  text = ""; 
        print, text = "smh26_x_mm_abs_active  =              0 ;";
        print,  text = ""; 
      }

      IF (dkf_x <> 0.) {  
        IF (wo_PRQDN90 == 1) {
           print,  text = "";   
           print,  text = "!------------------------------------------------------------------------------------------------------------";      
           print,  text = "! Tune knobs using the low energy quadrupoles without the PR.QDN90 (for LHC-type beams with long flat bottom)"; 
           print,  text = "!------------------------------------------------------------------------------------------------------------"; 
           print,  text = ""; 
           print,  text = "kPRQDN90 = 0.;"; 
           print,  text = "";      
           print,  text = "qx_leq             =       0.0000000000 ;";
           print,  text = "qy_leq             =       0.0000000000 ;";  
           print,  text = "";      
           printf, text = "dkf_x              = %F ;", value = dkf_x;
           printf, text = "dkd_x              = %F ;", value = dkd_x;
           print,  text = ""; 
           printf, text = "dkf_y              = %F ;", value = dkf_y;
           printf, text = "dkd_y              = %F ;", value = dkd_y;
           print,  text = "";           
           printf, text = "kf := %F + dkf_x * qx_leq + dkf_y * qy_leq;", value = kf_0;
           printf, text = "kd := %F + dkd_x * qx_leq + dkd_y * qy_leq;", value = kd_0;
           print,  text = "";
        } 
        ELSE {  
           print,  text = "";   
           print,  text = "!-------------------------------------------------";      
           print,  text = "! Tune knobs using the low energy quadrupoles"; 
           print,  text = "!-------------------------------------------------"; 
           print,  text = "";      
           print,  text = "kPRQDN90 := kd;"; 
           print,  text = "";  
           print,  text = "qx_leq             =       0.0000000000 ;";
           print,  text = "qy_leq             =       0.0000000000 ;";  
           print,  text = "";      
           printf, text = "dkf_x              = %F ;", value = dkf_x;
           printf, text = "dkd_x              = %F ;", value = dkd_x;
           print,  text = ""; 
           printf, text = "dkf_y              = %F ;", value = dkf_y;
           printf, text = "dkd_y              = %F ;", value = dkd_y;
           print,  text = "";           
           printf, text = "kf := %F + dkf_x * qx_leq + dkf_y * qy_leq;", value = kf_0;
           printf, text = "kd := %F + dkd_x * qx_leq + dkd_y * qy_leq;", value = kd_0;
           print,  text = "";
        }
      }

      IF (zero_disp_knobs <> 0.) {
           print,  text = "";   
           print,  text = "!-------------------------------------------------";      
           print,  text = "! Zero dispersion knobs"; 
           print,  text = "!-------------------------------------------------"; 
           print,  text = "";      
           print,  text = "Dx_bwsh54_leq    =       0.0000000000 ;";
           print,  text = "Dx_bwsh65_leq    =       0.0000000000 ;";  
           print,  text = "Dx_bwsh68_leq    =       0.0000000000 ;"; 
           print,  text = "Dx_bgi82_leq     =       0.0000000000 ;"; 
           print,  text = ""; 

           printf, text = "bwsh54_dkN05     = %F ;", value = bwsh54_dkN05;
           printf, text = "bwsh54_dkN21     = %F ;", value = bwsh54_dkN21;
           printf, text = "bwsh54_dkW27     = %F ;", value = bwsh54_dkW27;
           printf, text = "bwsh54_dkW28     = %F ;", value = bwsh54_dkW28;
           printf, text = "bwsh54_dkN35     = %F ;", value = bwsh54_dkN35;
           printf, text = "bwsh54_dkN36     = %F ;", value = bwsh54_dkN36;
           printf, text = "bwsh54_dkN45     = %F ;", value = bwsh54_dkN45;
           printf, text = "bwsh54_dkN55     = %F ;", value = bwsh54_dkN55;
           printf, text = "bwsh54_dkW56     = %F ;", value = bwsh54_dkW56;
           printf, text = "bwsh54_dkN71     = %F ;", value = bwsh54_dkN71;
           printf, text = "bwsh54_dkN72     = %F ;", value = bwsh54_dkN72;
           printf, text = "bwsh54_dkN81     = %F ;", value = bwsh54_dkN81;
           printf, text = "bwsh54_dkN89     = %F ;", value = bwsh54_dkN89;
           printf, text = "bwsh54_dkN95     = %F ;", value = bwsh54_dkN95;
           printf, text = "bwsh54_dkN96     = %F ;", value = bwsh54_dkN96;
           print,  text = "";
           printf, text = "bwsh65_dkN05     = %F ;", value = bwsh65_dkN05;
           printf, text = "bwsh65_dkN09     = %F ;", value = bwsh65_dkN09;
           printf, text = "bwsh65_dkW17     = %F ;", value = bwsh65_dkW17;
           printf, text = "bwsh65_dkN21     = %F ;", value = bwsh65_dkN21;
           printf, text = "bwsh65_dkW31     = %F ;", value = bwsh65_dkW31;
           printf, text = "bwsh65_dkN39     = %F ;", value = bwsh65_dkN39;
           printf, text = "bwsh65_dkN45     = %F ;", value = bwsh65_dkN45;
           printf, text = "bwsh65_dkN49     = %F ;", value = bwsh65_dkN49;
           printf, text = "bwsh65_dkN55     = %F ;", value = bwsh65_dkN55;
           printf, text = "bwsh65_dkN67     = %F ;", value = bwsh65_dkN67;
           printf, text = "bwsh65_dkN77     = %F ;", value = bwsh65_dkN77;
           printf, text = "bwsh65_dkN81     = %F ;", value = bwsh65_dkN81;
           printf, text = "bwsh65_dkN85     = %F ;", value = bwsh65_dkN85;
           printf, text = "bwsh65_dkN89     = %F ;", value = bwsh65_dkN89;
           printf, text = "bwsh65_dkN99     = %F ;", value = bwsh65_dkN99;
           print,  text = "";   
           printf, text = "bwsh68_dkN09     = %F ;", value = bwsh68_dkN09;
           printf, text = "bwsh68_dkW17     = %F ;", value = bwsh68_dkW17;
           printf, text = "bwsh68_dkW18     = %F ;", value = bwsh68_dkW18;
           printf, text = "bwsh68_dkW27     = %F ;", value = bwsh68_dkW27;
           printf, text = "bwsh68_dkN35     = %F ;", value = bwsh68_dkN35;
           printf, text = "bwsh68_dkN49     = %F ;", value = bwsh68_dkN49;
           printf, text = "bwsh68_dkN50     = %F ;", value = bwsh68_dkN50;
           printf, text = "bwsh68_dkW59     = %F ;", value = bwsh68_dkW59;
           printf, text = "bwsh68_dkN67     = %F ;", value = bwsh68_dkN67;
           printf, text = "bwsh68_dkN71     = %F ;", value = bwsh68_dkN71;
           printf, text = "bwsh68_dkN77     = %F ;", value = bwsh68_dkN77;
           printf, text = "bwsh68_dkN78     = %F ;", value = bwsh68_dkN78;
           printf, text = "bwsh68_dkN85     = %F ;", value = bwsh68_dkN85;
           printf, text = "bwsh68_dkN86     = %F ;", value = bwsh68_dkN86;
           printf, text = "bwsh68_dkN95     = %F ;", value = bwsh68_dkN95;
           print,  text = "";           
           printf, text = "bgi82_dkN09      = %F ;", value = bgi82_dkN09;
           printf, text = "bgi82_dkW17      = %F ;", value = bgi82_dkW17;
           printf, text = "bgi82_dkW31      = %F ;", value = bgi82_dkW31;
           printf, text = "bgi82_dkW32      = %F ;", value = bgi82_dkW32;
           printf, text = "bgi82_dkN39      = %F ;", value = bgi82_dkN39;
           printf, text = "bgi82_dkN40      = %F ;", value = bgi82_dkN40;
           printf, text = "bgi82_dkN49      = %F ;", value = bgi82_dkN49;
           printf, text = "bgi82_dkN55      = %F ;", value = bgi82_dkN55;
           printf, text = "bgi82_dkW56      = %F ;", value = bgi82_dkW56;
           printf, text = "bgi82_dkN71      = %F ;", value = bgi82_dkN71;
           printf, text = "bgi82_dkN72      = %F ;", value = bgi82_dkN72;
           printf, text = "bgi82_dkN81      = %F ;", value = bgi82_dkN81;
           printf, text = "bgi82_dkN85      = %F ;", value = bgi82_dkN85;
           printf, text = "bgi82_dkN99      = %F ;", value = bgi82_dkN99;
           printf, text = "bgi82_dkN00      = %F ;", value = bgi82_dkN00;
           print,  text = "";           
           print,  text = "kPRQFN05 := kf + bwsh54_dkN05 * Dx_bwsh54_leq + bwsh65_dkN05 * Dx_bwsh65_leq;"; 
           print,  text = "kPRQFN09 := kf + bwsh65_dkN09 * Dx_bwsh65_leq + bwsh68_dkN09 * Dx_bwsh68_leq + bgi82_dkN09 * Dx_bgi82_leq;"; 
           print,  text = "kPRQFW17 := kf + bwsh65_dkW17 * Dx_bwsh65_leq + bwsh68_dkW17 * Dx_bwsh68_leq + bgi82_dkW17 * Dx_bgi82_leq;"; 
           print,  text = "kPRQDW18 := kd + bwsh68_dkW18 * Dx_bwsh68_leq;"; 
           print,  text = "kPRQFN21 := kf + bwsh54_dkN21 * Dx_bwsh54_leq + bwsh65_dkN21 * Dx_bwsh65_leq;"; 
           print,  text = "kPRQFW27 := kf + bwsh54_dkW27 * Dx_bwsh54_leq + bwsh68_dkW27 * Dx_bwsh68_leq;"; 
           print,  text = "kPRQDW28 := kd + bwsh54_dkW28 * Dx_bwsh54_leq;"; 
           print,  text = "kPRQFW31 := kf + bwsh65_dkW31 * Dx_bwsh65_leq + bgi82_dkW31 * Dx_bgi82_leq;"; 
           print,  text = "kPRQDW32 := kd + bgi82_dkW32 * Dx_bgi82_leq;";
           print,  text = "kPRQFN35 := kf + bwsh54_dkN35 * Dx_bwsh54_leq + bwsh68_dkN35 * Dx_bwsh68_leq;"; 
           print,  text = "kPRQDN36 := kd + bwsh54_dkN36 * Dx_bwsh54_leq;"; 
           print,  text = "kPRQFN39 := kf + bwsh65_dkN39 * Dx_bwsh65_leq + bgi82_dkN39 * Dx_bgi82_leq;"; 
           print,  text = "kPRQDN40 := kd + bgi82_dkN40 * Dx_bgi82_leq;"; 
           print,  text = "kPRQFN45 := kf + bwsh54_dkN45 * Dx_bwsh54_leq + bwsh65_dkN45 * Dx_bwsh65_leq;"; 
           print,  text = "kPRQFN49 := kf + bwsh65_dkN49 * Dx_bwsh65_leq + bwsh68_dkN49 * Dx_bwsh68_leq + bgi82_dkN49 * Dx_bgi82_leq;"; 
           print,  text = "kPRQDN50 := kd + bwsh68_dkN50 * Dx_bwsh68_leq;"; 
           print,  text = "kPRQFN55 := kf + bwsh54_dkN55 * Dx_bwsh54_leq + bwsh65_dkN55 * Dx_bwsh65_leq + bgi82_dkN55 * Dx_bgi82_leq;"; 
           print,  text = "kPRQDW56 := kd + bwsh54_dkW56 * Dx_bwsh54_leq + bgi82_dkW56 * Dx_bgi82_leq;"; 
           print,  text = "kPRQFW59 := kf + bwsh68_dkW59 * Dx_bwsh68_leq;"; 
           print,  text = "kPRQFN67 := kf + bwsh65_dkN67 * Dx_bwsh65_leq + bwsh68_dkN67 * Dx_bwsh68_leq;"; 
           print,  text = "kPRQFN71 := kf + bwsh54_dkN71 * Dx_bwsh54_leq + bwsh68_dkN71 * Dx_bwsh68_leq + bgi82_dkN71 * Dx_bgi82_leq;"; 
           print,  text = "kPRQDN72 := kd + bwsh54_dkN72 * Dx_bwsh54_leq + bgi82_dkN72 * Dx_bgi82_leq;"; 
           print,  text = "kPRQFN77 := kf + bwsh65_dkN77 * Dx_bwsh65_leq + bwsh68_dkN77 * Dx_bwsh68_leq;"; 
           print,  text = "kPRQDN78 := kd + bwsh68_dkN78 * Dx_bwsh68_leq;"; 
           print,  text = "kPRQFN81 := kf + bwsh54_dkN81 * Dx_bwsh54_leq + bwsh65_dkN81 * Dx_bwsh65_leq + bgi82_dkN81 * Dx_bgi82_leq;";        
           print,  text = "kPRQFN85 := kf + bwsh65_dkN85 * Dx_bwsh65_leq + bwsh68_dkN85 * Dx_bwsh68_leq + bgi82_dkN85 * Dx_bgi82_leq;";        
           print,  text = "kPRQDN86 := kd + bwsh68_dkN86 * Dx_bwsh68_leq;"; 
           print,  text = "kPRQFN89 := kf + bwsh54_dkN89 * Dx_bwsh54_leq + bwsh65_dkN89 * Dx_bwsh65_leq;"; 
           print,  text = "kPRQFN95 := kf + bwsh54_dkN95 * Dx_bwsh54_leq + bwsh68_dkN95 * Dx_bwsh68_leq;"; 
           print,  text = "kPRQDN96 := kd + bwsh54_dkN96 * Dx_bwsh54_leq;"; 
           print,  text = "kPRQFN99 := kf + bwsh65_dkN99 * Dx_bwsh65_leq + bgi82_dkN99 * Dx_bgi82_leq;"; 
           print,  text = "kPRQDN00 := kd + bgi82_dkN00 * Dx_bgi82_leq;"; 
           print,  text = ""; 
      }

      IF (dkPEBSW14_x_abs <> 0.) {
        print,  text = "";   
        print,  text = "!-------------------------------------------------";      
        print,  text = "! Extraction bump"; 
        print,  text = "!-------------------------------------------------";
        print,  text = "";  
        print, text = "smh16_x_mm_abs_active  =                  1 ;";         
        printf, text = "smh16_x_mm_abs         = %F ;", value = smh16_x_mm_abs;
        print, text = "smh16_x_mm             =       0.0000000000 ;";
        print, text = "smh16_px_urad          =       0.0000000000 ;";
        print, text = "tps15_ortho_x_mm       =       0.0000000000 ;";    
        print, text = "smh16_ortho_x_mm       =       0.0000000000 ;";    
        print,  text = "";      
        printf, text = "dkPEBSW12_x_abs        = %F ;", value = dkPEBSW12_x_abs;
        printf, text = "dkPEBSW14_x_abs        = %F ;", value = dkPEBSW14_x_abs;
        printf, text = "dkPEBSW20_x_abs        = %F ;", value = dkPEBSW20_x_abs;
        printf, text = "dkPEBSW22_x_abs        = %F ;", value = dkPEBSW22_x_abs;
        print,  text = "";      
        printf, text = "dkPEBSW12_x            = %F ;", value = dkPEBSW12_x;
        printf, text = "dkPEBSW14_x            = %F ;", value = dkPEBSW14_x;
        printf, text = "dkPEBSW20_x            = %F ;", value = dkPEBSW20_x;
        printf, text = "dkPEBSW22_x            = %F ;", value = dkPEBSW22_x;
        print,  text = "";      
        printf, text = "dkPEBSW12_px           = %F ;", value = dkPEBSW12_px;
        printf, text = "dkPEBSW14_px           = %F ;", value = dkPEBSW14_px;
        printf, text = "dkPEBSW20_px           = %F ;", value = dkPEBSW20_px;
        printf, text = "dkPEBSW22_px           = %F ;", value = dkPEBSW22_px;
        print,  text = "";   
        printf, text = "dkPEBSW12_ortho_x15    = %F ;", value = dkPEBSW12_ortho_x15;
        printf, text = "dkPEBSW14_ortho_x15    = %F ;", value = dkPEBSW14_ortho_x15;
        printf, text = "dkPEBSW20_ortho_x15    = %F ;", value = dkPEBSW20_ortho_x15;
        printf, text = "dkPEBSW22_ortho_x15    = %F ;", value = dkPEBSW22_ortho_x15;
        print,  text = ""; 
        printf, text = "dkPEBSW12_ortho_x16    = %F ;", value = dkPEBSW12_ortho_x16;
        printf, text = "dkPEBSW14_ortho_x16    = %F ;", value = dkPEBSW14_ortho_x16;
        printf, text = "dkPEBSW20_ortho_x16    = %F ;", value = dkPEBSW20_ortho_x16;
        printf, text = "dkPEBSW22_ortho_x16    = %F ;", value = dkPEBSW22_ortho_x16;
        print,  text = "";          
        print, text = "kPEBSW12 := dkPEBSW12_x_abs * smh16_x_mm_abs + dkPEBSW12_x * smh16_x_mm + dkPEBSW12_px * smh16_px_urad + dkPEBSW12_ortho_x15 * tps15_ortho_x_mm + dkPEBSW12_ortho_x16 * smh16_ortho_x_mm;";
        print, text = "kPEBSW14 := dkPEBSW14_x_abs * smh16_x_mm_abs + dkPEBSW14_x * smh16_x_mm + dkPEBSW14_px * smh16_px_urad + dkPEBSW14_ortho_x15 * tps15_ortho_x_mm + dkPEBSW14_ortho_x16 * smh16_ortho_x_mm;";
        print, text = "kPEBSW20 := dkPEBSW20_x_abs * smh16_x_mm_abs + dkPEBSW20_x * smh16_x_mm + dkPEBSW20_px * smh16_px_urad + dkPEBSW20_ortho_x15 * tps15_ortho_x_mm + dkPEBSW20_ortho_x16 * smh16_ortho_x_mm;";
        print, text = "kPEBSW22 := dkPEBSW22_x_abs * smh16_x_mm_abs + dkPEBSW22_x * smh16_x_mm + dkPEBSW22_px * smh16_px_urad + dkPEBSW22_ortho_x15 * tps15_ortho_x_mm + dkPEBSW22_ortho_x16 * smh16_ortho_x_mm;";
        print,  text = "";  
      } 
      ELSEIF (dkpeseh23_x_abs <> 0.){
        print,  text = "";   
        print,  text = "!-------------------------------------------------";      
        print,  text = "! Extraction bump"; 
        print,  text = "!-------------------------------------------------";
        print,  text = ""; 
        print, text = "seh23_x_mm_abs_active  =              1 ;";
        print, text = "smh57_x_mm_abs_active  =              1 ;";
        print,  text = "";      
        printf, text = "seh23_x_mm_abs      = %F ;", value = x_SEH23*1e3;
        printf, text = "smh57_x_mm_abs      = %F ;", value = x_SMH57*1e3;
        print,  text = "";      
        printf, text = "dkPESEH23_x_abs     = %F ;", value = dkPESEH23_x_abs;
        printf, text = "dkPESMH57_x_abs     = %F ;", value = dkPESMH57_x_abs;
        print,  text = "";      
        print, text = "kPEBSW23 := dkPESEH23_x_abs * seh23_x_mm_abs;";
        print, text = "kPEBSW57 := dkPESMH57_x_abs * smh57_x_mm_abs;";
        print,  text = "";  
      }
      ELSE{
        print,  text = "";   
        print,  text = "!-------------------------------------------------";      
        print,  text = "! Extraction bump"; 
        print,  text = "!-------------------------------------------------";
        print,  text = "";  
        print, text = "smh16_x_mm_abs_active  =              0 ;";
        print, text = "seh23_x_mm_abs_active  =              1 ;";
        print, text = "seh23_x_mm_abs         =              0 ;";
        print, text = "smh57_x_mm_abs_active  =              1 ;";
        print, text = "smh57_x_mm_abs         =              0 ;";
        print,  text = "";   
      }

      IF (dkPRDHZOC05_SMH16_x <> 0.) {
        print,  text = "";   
        print,  text = "!-------------------------------------------------";      
        print,  text = "! High-energy corrector knobs"; 
        print,  text = "!-------------------------------------------------";
        print,  text = "";  
        print,  text = "dhzoc_smh16_x_mm     =                    0 ;";
        print,  text = "dhzoc_qsk19_x_mm     =                    0 ;";
        print,  text = "dhzoc_smh16_px_urad  =                    0 ;";
        print,  text = "dhzoc_qsk19_px_urad  =                    0 ;";
        print,  text = "";      
        printf, text = "dkPRDHZOC05_SMH16_x      = %F ;", value = dkPRDHZOC05_SMH16_x;
        printf, text = "dkPRDHZOC18_SMH16_x      = %F ;", value = dkPRDHZOC18_SMH16_x;
        printf, text = "dkPRDHZOC60_SMH16_x      = %F ;", value = dkPRDHZOC60_SMH16_x;
        print,  text = "";   
        printf, text = "dkPRDHZOC05_QSK19_x      = %F ;", value = dkPRDHZOC05_QSK19_x;
        printf, text = "dkPRDHZOC18_QSK19_x      = %F ;", value = dkPRDHZOC18_QSK19_x;
        printf, text = "dkPRDHZOC60_QSK19_x      = %F ;", value = dkPRDHZOC60_QSK19_x;
        print,  text = "";   
        printf, text = "dkPRDHZOC05_SMH16_px     = %F ;", value = dkPRDHZOC05_SMH16_px;
        printf, text = "dkPRDHZOC18_SMH16_px     = %F ;", value = dkPRDHZOC18_SMH16_px;
        printf, text = "dkPRDHZOC60_SMH16_px     = %F ;", value = dkPRDHZOC60_SMH16_px;
        print,  text = "";  
        printf, text = "dkPRDHZOC05_QSK19_px     = %F ;", value = dkPRDHZOC05_QSK19_px;
        printf, text = "dkPRDHZOC18_QSK19_px     = %F ;", value = dkPRDHZOC18_QSK19_px;
        printf, text = "dkPRDHZOC60_QSK19_px     = %F ;", value = dkPRDHZOC60_QSK19_px;
        print,  text = "";    
        print, text = "kPRDHZOC05 := dkPRDHZOC05_SMH16_x * dhzoc_smh16_x_mm + dkPRDHZOC05_QSK19_x * dhzoc_qsk19_x_mm + dkPRDHZOC05_SMH16_px * dhzoc_smh16_px_urad + dkPRDHZOC05_QSK19_px * dhzoc_qsk19_px_urad;";
        print, text = "kPRDHZOC18 := dkPRDHZOC18_SMH16_x * dhzoc_smh16_x_mm + dkPRDHZOC18_QSK19_x * dhzoc_qsk19_x_mm + dkPRDHZOC18_SMH16_px * dhzoc_smh16_px_urad + dkPRDHZOC18_QSK19_px * dhzoc_qsk19_px_urad;";
        print, text = "kPRDHZOC60 := dkPRDHZOC60_SMH16_x * dhzoc_smh16_x_mm + dkPRDHZOC60_QSK19_x * dhzoc_qsk19_x_mm + dkPRDHZOC60_SMH16_px * dhzoc_smh16_px_urad + dkPRDHZOC60_QSK19_px * dhzoc_qsk19_px_urad;";
        print,  text = "";  
      } 

      assign, echo = terminal;
    };


    ''')





import numpy as np
import os



def replace_extracted_val(monitor, delta=False, turn=None, turn_all_after=False, max_search=1e4, message=True):
    xcopy = monitor.x.copy()
    pxcopy = monitor.px.copy()
    if delta is True:
        deltacopy = monitor.delta.copy()


    if turn is not None:
        extracted_mask = xcopy[:, turn] == 0  # xcopy MUST EXISTS!
    else:  # inutile
        turn = len(xcopy[0, :]) - 1
        extracted_mask = xcopy[:, turn] == 0

    find_val = True
    search_counter = 1
    while find_val:
        if turn_all_after:  # This modifies all the values after some turn(param) to be as the last value before extraction
            xcopy[extracted_mask, turn:] = np.tile(xcopy[extracted_mask, turn - search_counter],
                                                   (len(monitor.x[0, :]) - turn, 1)).transpose()  #
            pxcopy[extracted_mask, turn:] = np.tile(pxcopy[extracted_mask, turn - search_counter],
                                                    (len(monitor.x[0, :]) - turn, 1)).transpose()
            if 'deltacopy' in locals():
                deltacopy[extracted_mask, turn:] = np.tile(deltacopy[extracted_mask, turn - search_counter],
                                                           (len(monitor.x[0, :]) - turn, 1)).transpose()
        else:
            xcopy[extracted_mask, turn] = xcopy[extracted_mask, turn - search_counter]
            pxcopy[extracted_mask, turn] = pxcopy[extracted_mask, turn - search_counter]
            if 'deltacopy' in locals():
                deltacopy[extracted_mask, turn] = deltacopy[extracted_mask, turn - search_counter]

        search_counter += 1
        extracted_mask = xcopy[:, turn] == 0
        if np.sum(extracted_mask) == 0:
            if message:
                print(f'process finished in {search_counter - 1} cycles')
            find_val = False

        if (turn - search_counter) < 0 & (turn != 0):
            if message:
                zero_first_turn = np.sum(monitor.x[:, 0] == 0)
                print(f'index exceeded ---> {zero_first_turn} particles are extracted the first turn ')
            find_val = False
            # break

        if search_counter > max_search:
            if message:
                print('maximum search iterations reached --> abort ')
            break
    if 'deltacopy' in locals():
        return xcopy, pxcopy, deltacopy
    else:
        return xcopy, pxcopy

def replace_array(arr, ind=False):
    arr_copy = arr.copy()
    index = np.where(arr == 0)[0][0]
    arr_copy[index:] = arr_copy[index - 1]
    if ind == False:
        return arr_copy
    else:
        return arr_copy, index


def replace_null(arr, ind=False):
    # WORKS WITH THE ENTIRE MATRIX
    # can return the index consistent with the particles.at_turn

    li = []
    index = np.zeros(shape=(arr.shape[0]))
    for i in range(arr.shape[0]):
        if np.sum(arr[i] == 0) != 0:
            if ind == True:
                replaced = replace_array(arr[i], ind=ind)
                index[i] = int(replaced[
                                   1])  # aggiunge alla particella numero i+1 l'indice in cui è stato trovato il primo zer, quidi in cui è stata estratta
                li.append(replaced[0])
            else:
                li.append(replace_array(arr[i]))
        else:
            li.append(arr[i])
    if ind == False:
        return np.array(li)
    else:
        return np.array(li), index



def _clean_dependencies(dependencies):
    clean_dep = []
    for x in dependencies:
        clean_dep.append(str(x)[6:].strip("']"))
    return clean_dep


def search_folder(directory, folder_name):
    """
    Search for a folder within a specified directory.

    Args:
        directory (str): The directory to search in.
        folder_name (str): The name of the folder to search for.

    Returns:
        str: The path to the found folder, or None if not found.
    """
    # Iterate through all items in the directory
    for root, dirs, files in os.walk(directory):
        # Check if the folder_name is in the list of directories
        if folder_name in dirs:
            # If found, return the full path to the folder
            return os.path.join(root, folder_name)
    # If not found, return None
    return None





def read_from_txt(filename,param,type_='float'):
    '''Search a parameters in the params file (Benchmark_dir)'''
    assert type_ in ['float','int','string']
    with open(filename) as f:
        lines = f.readlines()
        for line in lines:
            line = line.strip()
            if param in line:
                val = line[line.index('=')+1:].strip()
                if type_=='float':
                    return float(val)
                elif type_=='string':
                    return val # already a string
                else:
                    return int(val)









import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib import pyplot as plt
from matplotlib.animation import FuncAnimation
import os


from .utility import replace_extracted_val
from . import slowex_helper as slwex
from . import xsuite_helper as xh
from .machine_class import Rfko

#######################
##########         STATIC PLOT FUNCTIONS
def plot_extracted(out_dict, figsize=(10, 8), rf_plot=True,beam_loss=False):

    if isinstance(out_dict, Rfko):
        out_dict = out_dict.to_dict()

    particles = out_dict['particles']
    particles_lost = np.bincount(particles.at_turn)
    if rf_plot:
        #adding the option for the non linear chirp
        fig, ax = plt.subplots(2, 1, sharex=True, height_ratios=[6, 2], figsize=figsize)
        if beam_loss:
            ax[0].plot(np.arange(len(particles_lost) - 1),len(particles.x)- np.cumsum(particles_lost[:-1]))
        else:
            ax[0].plot(np.arange(len(particles_lost) - 1), np.cumsum(particles_lost[:-1]))
        _plot_signal(out_dict, ax=ax[1])
        fig.suptitle('Particle extraction vs rf signal')
        fig.supxlabel('Turns')
        ax[0].set_ylabel('Extracted particles')
        ax[1].set_ylabel('Frequencies')
    else:
        plt.figure(figsize=figsize)
        if beam_loss:
            plt.plot(np.arange(len(particles_lost) - 1), len(particles.x) -np.cumsum(particles_lost[:-1]))
        else:
            plt.plot(np.arange(len(particles_lost) - 1),np.cumsum(particles_lost[:-1]))
    return None

def plot_flex(out_dict, figsize=(12, 8), filename=None,message='',n_part='all', turn_range=None,s=2,particles_id=None):
    ''' DEFECT : I can't find a straigth forward way to include the limit of the number of particles also for the extracted
    this is related to the fact that there is not a correspondece from the particles object that store the last turn
     to the monitor that records them'''
    if isinstance(out_dict,Rfko):
        out_dict=out_dict.to_dict()


    monitor = out_dict['monitor']
    particles = out_dict['particles']
    if (n_part=='all')&(particles_id is None):
        ids = [x for x in range(out_dict['params']['n_part'])]
    elif particles_id is None:
        ids = np.random.randint(0,out_dict['params']['n_part'],n_part)
    else:
        ids = [x for x in monitor.particle_id[:,0] if x in particles_id ]
      #Similar version with Boolen indexing ------>
        # Tese are EQUIVALENT ONLY if the particles_id in the monitor are always SORTED
        #ids = np.isin(monitor.particle_id[:,0],particles_id)
    if turn_range is not None:
        try:
            #if turn range is only a number it's considered the maximum turn number to plot
            len(turn_range)
        except:
            turn_range = [turn_range]
        if len(turn_range) == 2:
            x = monitor.x[ids, turn_range[0]:turn_range[1]]
            px = monitor.px[ids, turn_range[0]:turn_range[1]]
            x_extr = particles.x
            px_extr = particles.px
            mask_extr = (particles.state == 0) & (particles.at_turn < turn_range[1]) & (
                        particles.at_turn >= turn_range[0])
        else:
            x = monitor.x[ids, :turn_range[0]]
            px = monitor.px[ids, :turn_range[0]]
            mask_extr = (particles.state == 0) & (particles.at_turn < turn_range[0])
            x_extr = particles.x
            px_extr = particles.px

    else:
        x = monitor.x[ids,:]
        px = monitor.px[ids,:]
        x_extr = particles.x
        px_extr = particles.px
        mask_extr = particles.state == 0
    if out_dict['params']['SEPTA_X_MM'] is not None:
        septa = out_dict['params']['SEPTA_X_MM']

    mask_mon = x != 0
    # fig,ax = plt.subplots(figsize=figsize)
    plt.figure(figsize=figsize)
    if turn_range is not None:
        color_array = np.arange(len(x[0, :])) + 1
    else:
        color_array = np.arange(out_dict['params'][
                                    'n_turns']) + 1  # to adjust the scales of particles color to be (1,n_turns)

    colors = np.tile(color_array, (len(ids), 1))  # Fix dimensions
    

    #### SET axes LIMITS
    axmargin = 0.05
    if out_dict['params']['SEPTA_X_MM'] is not None:
        xmin = np.min(np.concatenate((x_extr[mask_extr].flatten(),x[mask_mon],[septa / 1000])))
        xmax = np.max(np.concatenate((x_extr[mask_extr].flatten(),x[mask_mon],[septa / 1000])))
    else:
        xmin = np.min(np.concatenate((x_extr[mask_extr].flatten(), x[mask_mon])))
        xmax = np.max(np.concatenate((x_extr[mask_extr].flatten(), x[mask_mon])))

    ymin = np.min(np.concatenate((px_extr[mask_extr].flatten(),px[mask_mon])))
    ymax = np.max(np.concatenate((px_extr[mask_extr].flatten(),px[mask_mon])))
    plt.xlim(xmin - np.abs(xmin - xmax) *axmargin, xmax + np.abs(xmax-xmin) * axmargin)
    plt.ylim(ymin - np.abs(ymin - ymax) * axmargin, ymax + np.abs(ymax - ymin) * axmargin)

    #PLOT
    plt.scatter(x_extr[mask_extr], px_extr[mask_extr], c=particles.at_turn[mask_extr], marker='P', alpha=0.5,s=s)
    plt.scatter(x[mask_mon], px[mask_mon], c=colors[mask_mon], alpha=0.5,s=s)
    plt.colorbar(label='Turns')
    #### SEPTA PLOT
    if out_dict['params']['SEPTA_X_MM'] is not None:
        plt.axvline(septa / 1000, c="r", linestyle="--")
    if filename is not None:
        plt.savefig(filename+message)
    return None




def stein_static(out_dict, at_turn, rmv_disp=False):
    ''' Create a steinbach diagram at a specific turn'''
    if isinstance(out_dict, Rfko):
        out_dict = out_dict.to_dict()

    twiss = out_dict['twiss']
    line = out_dict['line']
    monitor = out_dict['monitor']
    print('Bakup')
    if hasattr(monitor,'placed_at_element'):
        at_element=monitor.placed_at_element
    else:
        at_element = 'pe.smh57'
    ### PROVO A CREARE UN DIAGRAMMA DI STEINBACH
    S_virtual, S_mu = slwex.Virtual_sextupole_complete(out_dict)
    ######## I have to remove the closed orbit to calculate the amplitudes
    xcopy, pxcopy, delta_all = replace_extracted_val(monitor, delta=True,
                                                     turn=at_turn)  ### processed to place the last measured position on each extracted particle
    # REMOVE CLOSED ORBIT
    x = xcopy[:, at_turn] - twiss.x[twiss.name==at_element]
    px = pxcopy[:, at_turn] - twiss.px[twiss.name==at_element]
    deltas = delta_all[:, at_turn]


    # DISTORTION CORR
    argmin = np.argmin(np.abs(twiss.mux * 2 * np.pi - S_mu))
    Xo = twiss.x[argmin] / np.sqrt(twiss.betx[argmin])
    # there should be the correction from the closed orbit
    Qx_real = twiss.qx - twiss.dqx * deltas - S_virtual*Xo/(2*np.pi)

    # to see it with removed dispersion
    if np.sum(deltas != 0):
        x = x - twiss.dx[twiss.name==at_element] * deltas
        px = px - twiss.dpx[twiss.name==at_element] * deltas

    # parameters for normalization
    betx_monitor = twiss.betx[twiss.name == at_element]
    alfx_monitor = twiss.alfx[twiss.name == at_element]
    Xn = x / np.sqrt(betx_monitor)
    Xpn = (alfx_monitor / np.sqrt(betx_monitor)) * x + np.sqrt(betx_monitor) * px

    amplitude_n = np.sqrt(Xn ** 2 + Xpn ** 2)

    ############# ADDING color for the visualization of the extracted particles
    extracted_mask = monitor.x[:, at_turn] == 0


    dQ = np.linspace(0, 0.06, 500)
    Ax = np.sqrt(48 * np.sqrt(3) * np.pi) * dQ / S_virtual

    plt.figure(figsize=(13, 8))
    plt.scatter(Qx_real, amplitude_n, s=1, c=extracted_mask)
    plt.scatter(dQ + 6 + 1 / 3, Ax, c='r', s=3)
    plt.scatter(6 + 1 / 3 - dQ, Ax, c='r', s=3)
    # plt.savefig(f'fig_wes/Steinbach_Sept_sept{SEPTA_X_MM}_dpp{DPP_FACTOR}_emitx{exn}_Gain{volt}_RF{rf10MHz_enable}')


##############
#        ANIMATIONS
#############

# SISTEMARE EFFETTO, DISPERSIONE, SUL TUNE DELLE PARTICELLE ESTRATTE E SISTEMARE LA NECESSITà DEL FILENAME

def anim_steinbach_(out_dict, filename=None,message='', step=5, SEPTA=None):
    if isinstance(out_dict, Rfko):
        out_dict = out_dict.to_dict()

    # TO SE THE CORRECTION OF TUNE DUE TO DISTORTION
    particles = out_dict['particles']
    monitor = out_dict['monitor']
    twiss = out_dict['twiss']
    #line_df = out_dict['line'].to_pandas()
    data = {
        'at_turn': particles.at_turn,
        'x': particles.x,
        'px': particles.px,
        'delta': particles.delta,
    }
    df = pd.DataFrame(data)
    # extracting the dataframe form, easier to work with
    twiss_df = twiss.to_pandas()
    # Virtual sextupole
    S_virtual, _ = slwex.Virtual_sextupole_complete(out_dict)
    # Resonance line plot
    dQ = np.linspace(0, 0.06, 500) #
    if S_virtual != 0:
        Ax = np.sqrt(48 * np.sqrt(3) * np.pi) * dQ / S_virtual
    else:
        Ax = np.zeros(dQ.shape)

    n_turns = out_dict['params']['n_turns']
    #### figurw init
    fig, ax = plt.subplots(figsize=(12, 8))
    scat = ax.scatter([], [], s=1)

    # Calculating beta and alf for being able to normalize
    betx_monitor = twiss_df[twiss_df['name'] == 'pe.smh57'].betx.iloc[0]
    alfx_monitor = twiss_df[twiss_df['name'] == 'pe.smh57'].alfx.iloc[0]

    xcopy = monitor.x
    pxcopy = monitor.px
    deltas_corr = monitor.delta
    

    def update(frame):
        # a frame is a turn
        at_turn = frame
        if frame%50==0:
            print(f'frame {frame}')
        ax.clear()
        # PLOT RESONANCE LINE
        ax.plot(dQ + 6 + 1 / 3, Ax, c='r')
        ax.plot(6 + 1 / 3 - dQ, Ax, c='r')

        # extracted particles
        extracted_mask = monitor.x[:, at_turn] == 0
        x_extr = df[df['at_turn'] <= at_turn]['x']
        px_extr = df[df['at_turn'] <= at_turn]['px']
        deltas_extr = df[df['at_turn']<=at_turn]['delta']
        # x,px are values at_turn=frame
        # REMOVE CLOSED ORBIT
        x = xcopy[np.logical_not(extracted_mask), at_turn] - twiss_df[twiss_df['name'] == 'pe.smh57'].x.iloc[0]
        px = pxcopy[np.logical_not(extracted_mask), at_turn] - twiss_df[twiss_df['name'] == 'pe.smh57'].px.iloc[0]
        x_extr = x_extr - twiss_df[twiss_df['name'] == 'pe.smh57'].x.iloc[0]
        px_extr = px_extr - twiss_df[twiss_df['name'] == 'pe.smh57'].px.iloc[0]

        # CORRECTION OF THE DISPERSIONAL EFFECTS||
        if np.sum(monitor.delta.flatten()) != 0:
            x = x - twiss_df[twiss_df['name'] == 'pe.smh57'].dx.iloc[0] * deltas_corr[np.logical_not(extracted_mask), at_turn]
            px = px - twiss_df[twiss_df['name'] == 'pe.smh57'].dpx.iloc[0] * deltas_corr[np.logical_not(extracted_mask), at_turn]
            x_extr = x_extr - twiss_df[twiss_df['name'] == 'pe.smh57'].dx.iloc[0] * deltas_extr
            px_extr = px_extr - twiss_df[twiss_df['name'] == 'pe.smh57'].dpx.iloc[0] * deltas_extr
        # NORMALIZE
        Xn = x / np.sqrt(betx_monitor)
        Xpn = (alfx_monitor / np.sqrt(betx_monitor)) * x + np.sqrt(betx_monitor) * px
        X_extr = x_extr / np.sqrt(betx_monitor)
        Px_extr = x_extr * alfx_monitor / np.sqrt(betx_monitor) + px_extr * np.sqrt(betx_monitor)
        # AMPLITUDE
        amplitude_n = np.sqrt(Xn ** 2 + Xpn ** 2)
        amplitude_extr = np.sqrt(X_extr ** 2 + Px_extr ** 2)

        # TUNE--- no distortion correction because it causes problems--->> as the commented block
        Qx_real = twiss.qx + twiss.dqx * deltas_corr[np.logical_not(extracted_mask), at_turn]

        # Qx_real = twiss.qx - twiss.dqx * deltas_corr[np.logical_not(extracted_mask), at_turn] - \
        #         twiss_df[twiss_df['name'] == 'pe.smh57'].x.iloc[
        #            0] * S_virtual / (4 * np.pi * np.sqrt(betx_monitor))

        ## LIMS
        axs_margins = 0.1
        xmin = np.min(np.concatenate((Qx_real,6+1/3-dQ[0])))
        xmax = np.max(np.concatenate( (Qx_real,6+1/3+dQ[-1]) ))
        ymin = np.min(np.concatenate((amplitude_extr,amplitude_n,Ax)) )
        ymax = np.max(np.concatenate((amplitude_extr,amplitude_n,Ax)) )
        ax.set_xlim(xmin-(xmax-xmin)*axs_margins,xmax+(xmax-xmin)*axs_margins)
        ax.set_ylim(ymin -(ymax-ymin)*axs_margins,ymax+(ymax-ymin)*axs_margins)
        Qx_extr = twiss.qx + twiss.dqx * deltas_extr
        ax.set_title(f'Steinbach turn :{frame}/{n_turns}')
        ax.set_xlabel('Amplitude')
        ax.set_ylabel('Tune')
        ax.scatter(Qx_real, amplitude_n, s=1, c='b')
        ax.scatter(Qx_extr, amplitude_extr, s=1, c='g')



        return scat,

    ani = FuncAnimation(fig, update, frames=range(0, n_turns, step), blit=True)
    if filename is None:
        filename = f'part{out_dict["params"]["n_part"]}_DPP{out_dict["params"]["DPP_FACTOR"]}_Volt{out_dict["params"]["volt"]}'

    # plt.show()
    if os.path.isdir('./gif_wes'):
        ani.save(f'./gif_wes/steinbach_{message}_{filename}.gif', dpi=200)
    else:
        ani.save(f'./steinbach_{message}_{filename}.gif', dpi=200)


def anim_steinbach(out_dict, filename=None, message='', step=5, SEPTA=None):
    if isinstance(out_dict, Rfko):
        out_dict = out_dict.to_dict()

    ######
    ######### CREATING SOME VARIABLES
    particles = out_dict['particles']
    monitor = out_dict['monitor']
    twiss = out_dict['twiss']
    # extracting the dataframe form, easier to work with
    twiss_df = twiss.to_pandas()
    n_turns = out_dict['params']['n_turns']

    data = {
        'at_turn': particles.at_turn,
        'x': particles.x,
        'px': particles.px,
        'delta': particles.delta,
    }
    df = pd.DataFrame(data)

    # CALCULATION OF Virtual sextupole
    S_virtual, S_mu = slwex.Virtual_sextupole_complete(out_dict)

    # Resonance line plot
    dQ = np.linspace(0, 0.06, 500)  #
    if S_virtual != 0:
        Ax = np.sqrt(48 * np.sqrt(3) * np.pi) * dQ / S_virtual
    else:
        Ax = np.zeros(dQ.shape)

    #### figure init
    fig, ax = plt.subplots(figsize=(12, 8))
    scat = ax.scatter([], [], s=1)

    # Calculating beta and alf for being able to normalize
    betx_monitor = twiss_df[twiss_df['name'] == 'pe.smh57'].betx.iloc[0]
    alfx_monitor = twiss_df[twiss_df['name'] == 'pe.smh57'].alfx.iloc[0]

    xcopy = monitor.x
    pxcopy = monitor.px
    deltas_corr = monitor.delta

    def update(frame):
        # a frame is a turn
        at_turn = frame
        if frame % 50 == 0:
            print(f'frame {frame}')
        ax.clear()
        # PLOT RESONANCE LINE
        ax.plot(dQ + 6 + 1 / 3, Ax, c='r')
        ax.plot(6 + 1 / 3 - dQ, Ax, c='r')

        # mask extracted particles of the monitor --> 0 coordinates
        extracted_mask = monitor.x[:, at_turn] == 0
        # extracted particles position AT TURN, they retain the coordinates
        x_extr = df[df['at_turn'] <= at_turn]['x']
        px_extr = df[df['at_turn'] <= at_turn]['px']
        deltas_extr = df[df['at_turn'] <= at_turn]['delta']
       

        # coordinate at turn + REMOVE CLOSED ORBIT
        x = xcopy[np.logical_not(extracted_mask), at_turn] - twiss_df[twiss_df['name'] == 'pe.smh57'].x.iloc[0]
        px = pxcopy[np.logical_not(extracted_mask), at_turn] - twiss_df[twiss_df['name'] == 'pe.smh57'].px.iloc[0]
        x_extr = x_extr - twiss_df[twiss_df['name'] == 'pe.smh57'].x.iloc[0]
        px_extr = px_extr - twiss_df[twiss_df['name'] == 'pe.smh57'].px.iloc[0]

        # REMOVE THE DISPERSION EFFECT||
        if np.sum(monitor.delta.flatten()) != 0:
            x = x - twiss_df[twiss_df['name'] == 'pe.smh57'].dx.iloc[0] * deltas_corr[
                np.logical_not(extracted_mask), at_turn]
            px = px - twiss_df[twiss_df['name'] == 'pe.smh57'].dpx.iloc[0] * deltas_corr[
                np.logical_not(extracted_mask), at_turn]
            x_extr = x_extr - twiss_df[twiss_df['name'] == 'pe.smh57'].dx.iloc[0] * deltas_extr
            px_extr = px_extr - twiss_df[twiss_df['name'] == 'pe.smh57'].dpx.iloc[0] * deltas_extr

        # NORMALIZE
        Xn = x / np.sqrt(betx_monitor)
        Xpn = (alfx_monitor / np.sqrt(betx_monitor)) * x + np.sqrt(betx_monitor) * px
        X_extr = x_extr / np.sqrt(betx_monitor)
        Px_extr = x_extr * alfx_monitor / np.sqrt(betx_monitor) + px_extr * np.sqrt(betx_monitor)
        # AMPLITUDE
        amplitude_n = np.sqrt(Xn ** 2 + Xpn ** 2)
        amplitude_extr = np.sqrt(X_extr ** 2 + Px_extr ** 2)

        # TUNE CORRECTIONS commented block
        Xco = twiss_df.x[np.argmin(np.abs(twiss_df.mux * 2 * np.pi - S_mu))] / np.sqrt(betx_monitor)
        Qx_real = twiss.qx + twiss.dqx * deltas_corr[np.logical_not(extracted_mask), at_turn] - S_virtual * Xco / (
                    np.pi * 2)
        Qx_extr = twiss.qx + twiss.dqx * deltas_extr - S_virtual * Xco / (np.pi * 2)
        ## LIMS

        ax.set_xlim(6.25, 6.4)
        # ax.set_ylim()

        ax.set_title(f'Steinbach turn :{frame}/{n_turns}')
        ax.set_xlabel('Amplitude')
        ax.set_ylabel('Tune')
        ax.scatter(Qx_real, amplitude_n, s=1, c='b')
        ax.scatter(Qx_extr, amplitude_extr, s=1, c='g')

        return scat,

    ani = FuncAnimation(fig, update, frames=range(0, n_turns, step), blit=True)
    if filename is None:
        filename = f'part{out_dict["params"]["n_part"]}_DPP{out_dict["params"]["DPP_FACTOR"]}_Volt{out_dict["params"]["volt"]}'

    # plt.show()
    if os.path.isdir('./gif_wes'):
        ani.save(f'./gif_wes/steinbach_{message}_{filename}.gif', dpi=200)
    else:
        ani.save(f'./steinbach_{message}_{filename}.gif', dpi=200)


def anim_steinbach2(out_dict, filename=None, message='', step=5, SEPTA=None):
    ''' Attempt to fo it better, and to have also the adaptive limits for the plot'''
    if isinstance(out_dict, Rfko):
        out_dict = out_dict.to_dict()

    # TO SE THE CORRECTION OF TUNE DUE TO DISTORTION
    particles = out_dict['particles']
    monitor = out_dict['monitor']
    twiss = out_dict['twiss']
    # line_df = out_dict['line'].to_pandas()
    data = {
        'at_turn': particles.at_turn,
        'x': particles.x,
        'px': particles.px,
        'delta': particles.delta,
    }
    df = pd.DataFrame(data)
    # extracting the dataframe form, easier to work with
    twiss_df = twiss.to_pandas()
    # FIXED PLOT
    # Virtual sextupole
    S_virtual, _ = slwex.Virtual_sextupole_complete(out_dict)
    # Resonance line
    dQ = np.linspace(0, 0.06, 500)
    if S_virtual != 0:
        Ax = np.sqrt(48 * np.sqrt(3) * np.pi) * dQ / S_virtual
    else:
        Ax = np.zeros(dQ.shape)

    n_turns = out_dict['params']['n_turns']
    #### figurw init
    fig, ax = plt.subplots(figsize=(12, 8))
    scat = ax.scatter([], [], s=1)

    # Calculating beta and alf for being able to normalize
    betx_monitor = twiss_df[twiss_df['name'] == 'pe.smh57'].betx.iloc[0]
    alfx_monitor = twiss_df[twiss_df['name'] == 'pe.smh57'].alfx.iloc[0]


    xcopy = monitor.x
    pxcopy = monitor.px
    deltas_corr = monitor.delta
    X,PX,X_p,PX_p = xh.Normalize_monitor(out_dict,normalize_particles=True)




    def update(frame):
        # a frame is a turn
        at_turn = frame
        if frame % 50 == 0:
            print(f'frame {frame}')
        ax.clear()
        # PLOT RESONANCE LINE
        ax.plot(dQ + 6 + 1 / 3, Ax, c='r')
        ax.plot(6 + 1 / 3 - dQ, Ax, c='r')

        # extracted particles
        extracted_mask = monitor.x[:, at_turn] == 0
        print(df['at_turn'] <= at_turn)
        X_extr = X_p[df['at_turn'] <= at_turn]
        PX_extr = PX_p[df['at_turn'] <= at_turn]
        deltas_extr = df[df['at_turn'] <= at_turn]['delta']

        X_turn = X[:,at_turn]
        PX_turn = PX[:,at_turn]


        # CORRECTION OF THE DISPERSIONAL EFFECTS||
        if np.sum(monitor.delta.flatten()) != 0:
            corrx = (monitor.delta[:,at_turn]*twiss_df[twiss_df.name=='pe.smh57'].dx)/np.sqrt(betx_monitor)
            corrpx = monitor.delta[:,at_turn]*twiss_df[twiss_df.name=='pe.smh57'].dpx*np.sqrt(betx_monitor)
            X_turn -= corrx
            PX_turn-= corrx*alfx_monitor+corrpx
            X_extr -= corrx
            PX_extr -= corrx*alfx_monitor+corrpx



        # AMPLITUDE
        amplitude_n = np.sqrt(X_turn ** 2 + PX_turn ** 2)
        amplitude_extr = np.sqrt(X_extr ** 2 + PX_extr ** 2)

        # TUNE--- no distortion correction because it causes problems--->> as the commented block
        Qx_real = twiss.qx + twiss.dqx * deltas_corr[np.logical_not(extracted_mask), at_turn]

        # Qx_real = twiss.qx - twiss.dqx * deltas_corr[np.logical_not(extracted_mask), at_turn] - \
        #         twiss_df[twiss_df['name'] == 'pe.smh57'].x.iloc[
        #            0] * S_virtual / (4 * np.pi * np.sqrt(betx_monitor))

        ## LIMS
        axs_margins = 0.1
        xmin = np.min(np.concatenate((Qx_real, 6 + 1 / 3 - dQ[0])))
        xmax = np.max(np.concatenate((Qx_real, 6 + 1 / 3 + dQ[-1])))
        ymin = np.min(np.concatenate((np.sqrt(X_extr ** 2 + PX_extr ** 2).flatten(), np.sqrt((X-corrx) ** 2 + (PX-corrx*alfx_monitor+corrpx) ** 2), Ax)))
        ymax = np.max(np.concatenate((np.sqrt(X_extr ** 2 + PX_extr ** 2).flatten(), np.sqrt((X-corrx) ** 2 + (PX-corrx*alfx_monitor+corrpx) ** 2), Ax)))
        ax.set_xlim(xmin - (xmax - xmin) * axs_margins, xmax + (xmax - xmin) * axs_margins)
        ax.set_ylim(ymin - (ymax - ymin) * axs_margins, ymax + (ymax - ymin) * axs_margins)
        Qx_extr = twiss.qx + twiss.dqx * deltas_extr
        ax.set_title(f'Steinbach turn :{frame}/{n_turns}')
        ax.set_xlabel('Amplitude')
        ax.set_ylabel('Tune')
        ax.scatter(Qx_real, amplitude_n, s=1, c='b')
        ax.scatter(Qx_extr, amplitude_extr, s=1, c='g')

        return scat,

    ani = FuncAnimation(fig, update, frames=range(0, n_turns, step), blit=True)
    if filename is None:
        filename = f'part{out_dict["params"]["n_part"]}_DPP{out_dict["params"]["DPP_FACTOR"]}_Volt{out_dict["params"]["volt"]}'

    # plt.show()
    if os.path.isdir('./gif_wes'):
        ani.save(f'./gif_wes/steinbach_{message}_{filename}.gif', dpi=200)
    else:
        ani.save(f'./steinbach_{message}_{filename}.gif', dpi=200)


def anim_ph_space(out_dict, filename=None,message='', dispersion=False, step=5):
    '''  MAKE ANIMATION OF THE PHASE SPACE AT SEPTA NON-NORMALIZED,
    if dispersion == True it also plots the phase space shifted for the dispersional effect
    '''
    if isinstance(out_dict, Rfko):
        out_dict = out_dict.to_dict()

    particles = out_dict['particles']
    monitor = out_dict['monitor']
    SEPTA_X_MM = out_dict['params']['SEPTA_X_MM'] 

    # UTILE PER EVITARE GROSSI BUG CREDO
    data = {
        'at_turn': particles.at_turn,
        'x': particles.x,
        'px': particles.px,
        'delta': particles.delta,
    }
    n_turns = out_dict['params']['n_turns']
    df = pd.DataFrame(data)
    fig, ax = plt.subplots(figsize=(8, 6))
    scat = ax.scatter([], [], s=2)

    def update(frame):
        # a frame is a turn
        at_turn = frame
        if frame%50==0:
            print(f'frame {frame}')
        ax.clear()
        # position of EXTRACTED particles
        x_extr = df[df['at_turn'] <= at_turn]['x']  #
        px_extr = df[df['at_turn'] <= at_turn]['px']  #
        delta_extr = df[df['at_turn']<=at_turn]['delta']
        mask = monitor.x[:, at_turn] != 0  # mask on the monitor array, for the extracted particles (at_turn)
        ## particles position
        x_pres = monitor.x[mask, at_turn]
        px_pres = monitor.px[mask, at_turn]

        # TOLGO LA DISPERSIONE
        if dispersion:
            twiss_df = out_dict['line'].twiss(method='4d').to_pandas()
            x_pres_nodisp = x_pres - monitor.delta[mask, at_turn] * twiss_df[twiss_df.name == 'pe.smh57'].dx.iloc[0]
            px_pres_nodisp = px_pres - monitor.delta[mask, at_turn] * twiss_df[twiss_df.name == 'pe.smh57'].dpx.iloc[0]
            x_extr_nodisp = x_extr-delta_extr * twiss_df[twiss_df.name == 'pe.smh57'].dx.iloc[0]
            px_extr_nodisp = px_extr-delta_extr * twiss_df[twiss_df.name == 'pe.smh57'].dpx.iloc[0]
        ax.text(0.1, 0.9, f'{len(x_extr)} extracted', transform=ax.transAxes)
        ax.set_title(f'Phase space turn :{frame}/{n_turns}')
        ax.set_xlabel('x')
        ax.set_ylabel('px')
        axmargin = 0.05
        if SEPTA_X_MM is not None:
            ax.axvline(SEPTA_X_MM / 1000, c="r", linestyle="--")

        # XAXES LIMITS
        if SEPTA_X_MM is not None: # NOT GENERAL TO SET ONLY THE MINIMUM RESPECTIVE TO THE SEPTA
            xmin = np.min(np.concatenate((monitor.x.flatten(), np.array([SEPTA_X_MM / 1000]),particles.x.flatten())))
            xmax = np.max(np.concatenate((monitor.x.flatten(), np.array([SEPTA_X_MM / 1000]),particles.x.flatten())))
        else:
            xmin = np.min(monitor.x.flatten())
            xmax = np.max(monitor.x.flatten())
        ax.set_xlim(xmin - np.abs(xmin - xmax) *axmargin, xmax + np.abs(xmax-xmin) * axmargin)
        # YLIMITS
        ymin = np.min(monitor.px.flatten())
        ymax = np.max(monitor.px.flatten())
        ax.set_ylim(ymin - np.abs(ymin-ymax) * axmargin, ymax + np.abs(ymax-ymin) * axmargin)

        ax.scatter(x_extr, px_extr, c='g', s=2)
        ax.scatter(x_pres, px_pres, c='b', s=2, label='dispersion')
        if dispersion:
            ax.scatter(x_pres_nodisp, px_pres_nodisp, c='r', s=1, alpha=0.5, label='no_dispersion')
            ax.scatter(x_extr_nodisp,px_extr_nodisp,c='c',s=1,alpha=0.5)
            ax.legend(loc='lower left')

        return scat,

    ani = FuncAnimation(fig, update, frames=range(0, n_turns, step), blit=True)
    if filename is None:
        try:
            filename = f'n_part{out_dict["params"]["n_part"]}_QXtarget{out_dict["params"]["QX_TARGET"]}_Volt{out_dict["params"]["volt"]}_duration{out_dict["params"]["duration"]}'
        except:
            filename = f'n_part{out_dict["params"]["n_part"]}_QXtarget{out_dict["twiss"].qx}_Volt{out_dict["params"]["volt"]}_duration{out_dict["params"]["duration"]}'

    if os.path.isdir('./gif_wes'):
        ani.save(f'./gif_wes/ph_space_{message}_{filename}.gif', dpi=200)
    else:
        ani.save(f'./gif_wes/ph_space_{message}_{filename}.gif', dpi=200)

    return None


def _plot_signal(out_dict, ax=None):
    # it works, it has to be improved, but it ok
    ### for now I don't care the sampling frequency, it doesn need to be representative in this respect
    f_rev = 1 / out_dict['twiss'].T_rev0  # better than using the one saved in the params, cause maybe the twiss is more updated
    f_start = out_dict['params']['freq_start_ratio'] * f_rev
    f_stop = out_dict['params']['freq_end_ratio'] * f_rev
    n_turn = out_dict['params']['n_turns']
    if out_dict['params']['chirp_type']=='linear':
        sim_time = n_turn / f_rev
        duration = out_dict['params']['duration']
        rep_time = sim_time / duration
        signal_f = np.tile(np.linspace(f_start, f_stop, 1000), round(rep_time))  # UN IDEA PER RIPETERE I LSEGNALE
        turns = np.linspace(0, n_turn, len(signal_f))
        if ax is None:
            plt.plot(turns, signal_f)
        else:
            ax.plot(turns, signal_f)
    else:
        sim_time = n_turn / f_rev
        duration = out_dict['params']['duration']
        rep_time = sim_time / duration
        signal_nonlin = np.concatenate((np.linspace(f_start, f_stop, 500),np.zeros(500)))
        signal_f = np.tile(signal_nonlin, round(rep_time))  # UN IDEA PER RIPETERE I LSEGNALE
        turns = np.linspace(0, n_turn, len(signal_f))
        if ax is None:
            plt.plot(turns, signal_f)
        else:
            ax.plot(turns, signal_f)


        return None



# Specifc internal functions
from .line_setup import build_line_baremachine
from .machine_class import  Rfko
from .quad_ripples import tracking as qr_tracking
from .xsuite_helper import tracking
from  .plotting1 import plot_extracted
# Modules import
from . import plotting1
from . import plotting2
from . import xsuite_actions as xa
from . import xsuite_helper as xh
from . import slowex_helper as slwex
from . import phy_helper as ph
from . import slowex_actions as slwexa
from . import check_functions as ch
from . import line_setup

# # if I do import package-name * it will import everything in this list
# __all__ = ['build_line_baremachine','tracking','builder_wrapper','plot_flex','plot_extracted']







# Import standard packages
import os
import sys
import numpy as np
import scipy
import logging

# Import simulation packages
import xtrack as xt
import xpart as xp
import xobjects as xo
from cpymad.madx import Madx
# CUSTOM functions there were previously on this notebook

try:
    import requests
except:
    print('the module request is not imported for a compatibility problem with OpeSSL.')
    print('this means that you cannot use the function rfko_setup with the corr_seq parameter set to False (True is the default)')




from . import phy_helper as ph
from . import slowex_helper as slwex

np.set_printoptions(threshold=sys.maxsize)  # To print tthe full length of np array

#logging.disable(logging.INFO)
#
#logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger(__name__) # I could use a diffrent name?


def set_logging(level):
    ''' St the level of logging for the logging in stance of the whole module'''
    if level in ['INFO', 'DEBUG', 'WARNING', 'ERROR', 'CRITICAL']:
        logger.setLevel(level=level)
    else:
        logger.error("Non recognized error")
        logger.debug(f"Use one of the following : {['INFO', 'DEBUG', 'WARNING', 'ERROR', 'CRITICAL']}")
        return None


def tune_match(mad, Qx, Qxp, Qy, Qyp,verbose=True):
    mad.input('''
    ptc_twiss_macro(order, dp, slice_flag): macro = {
    ptc_create_universe;
    ptc_create_layout, time=false, model=2, exact=true, method=6, nst=3;
    IF (slice_flag == 1){
        select, flag=ptc_twiss, clear;
        select, flag=ptc_twiss, column=name,keyword,s,l,x,px,beta11,beta22,disp1,k1l;
        ptc_twiss, closed_orbit, icase=56, no=order, deltap=dp, table=ptc_twiss, summary_table=ptc_twiss_summary, slice_magnets=true;
    }
    ELSE{
        select, flag=ptc_twiss, clear;
        select, flag=ptc_twiss, column=name,keyword,s,x,px,beta11,alfa11,beta22,alfa22,disp1,disp2,mu1,mu2,energy,l,angle,K1L,K2L,K3L,HKICK,SLOT_ID;    
        ptc_twiss, closed_orbit, icase=56, no=order, deltap=dp, table=ptc_twiss, summary_table=ptc_twiss_summary, normal;
    }
    ptc_end;
    };

    ''')

    # /**********************************************************************************
    # *                        Matching using the PFW
    # ***********************************************************************************/
    mad.input("Qx   := " + str(Qx) + "; !Horizontal Tune")
    mad.input("Qxp  := " + str(Qxp) + "; !Horizontal Chromaticity")

    mad.input("Qy   := " + str(Qy) + "; !Vertical Tune")
    mad.input("Qyp  := " + str(Qyp) + "; !Vertical Chromaticity")

    mad.input('''
    use, sequence=PS;
    match, use_macro;
            vary, name = k1prpfwf;
            vary, name = k1prpfwd;
            vary, name = k2prpfwf;
            vary, name = k2prpfwd;
            use_macro, name = ptc_twiss_macro(2,0,0);
            constraint, expr = table(ptc_twiss_summary,Q1)  = Qx;
            constraint, expr = table(ptc_twiss_summary,Q2)  = Qy;
            constraint, expr = table(ptc_twiss_summary,DQ1) = Qxp;
            constraint, expr = table(ptc_twiss_summary,DQ2) = Qyp;
    jacobian,calls=50000,bisec=3;
    ENDMATCH;
    ''')

    mad.use(sequence="PS")
    twiss_tune_matching = mad.twiss().dframe()  # Needed to refresh the tune values
    mad.input('qx = table(SUMM, Q1);')
    mad.input('qy = table(SUMM, Q2);')
    mad.input('qxp = table(SUMM, DQ1);')
    mad.input('qyp = table(SUMM, DQ2);')


    tune_info_dict = {"Qx": mad.globals["qx"], "Qy": mad.globals["qy"], "Qxp": mad.globals["qxp"],
                      "Qyp": mad.globals["qyp"]}
    pfw_dict = {"k1prpfwf": mad.globals["k1prpfwf"], "k1prpfwd": mad.globals["k1prpfwd"],
                "k2prpfwf": mad.globals["k2prpfwf"], "k2prpfwd": mad.globals["k2prpfwd"]}
    if verbose:
        print(
        f"H-tune: {round(tune_info_dict['Qx'], 3)}, H-Chroma: {round(tune_info_dict['Qxp'], 3)}\nV-Tune: {round(tune_info_dict['Qy'], 3)}, V-Chroma: {round(tune_info_dict['Qyp'], 3)}")
        print(f"")
        print(
        f"PFW settings: \n  k1prpfwf: {round(pfw_dict['k1prpfwf'], 6)}\n  k1prpfwd: {round(pfw_dict['k1prpfwd'], 6)}\n  k2prpfwf: {round(pfw_dict['k2prpfwf'], 6)}\n  k2prpfwd: {round(pfw_dict['k2prpfwd'], 6)}")

    return pfw_dict, tune_info_dict


def rfko_setup(REL_GAMMA=None, E_0 =193.766308,E_n =None,exn=2e-6, eyn=2e-6, duration=1 / 1000, volt=0, time=0.004, n_part=1000,
               n_turns=1000, freq_start_ratio=90 / 100, freq_end_ratio=110 / 100, sampling_freq=1e9,
               QY_TARGET=0.256, QPY_TARGET=-0.242, QX_TARGET=0.325, QPX_TARGET=-0.5, rf10MHz_enable='0',
               DPP_FACTOR=0, SEPTA_X_MM=-60, ctx=xo.ContextCpu(), chirp_type='linear', verbose=True, norm_gain=False,
               corr_seq=True, return_form='dict'):
    # Beam parameters

    every = ['freq_start_ratio', 'freq_end_ratio', 'duration', 'sampling_freq', 'REL_BETA', 'REL_GAMMA', 'exn',
             'eyn', 'p', 'Brho', 'QX_TARGET', 'QPX_TARGET', 'QY_TARGET', 'QPY_TARGET', 'volt', 'rf10MHz_enable', 'time',
             'n_part', 'n_turns', 'SEPTA_X_MM']
    if (E_n is None) & (REL_GAMMA is None):
        E_n = 0.650 # 650 Mev per nucleon


    E_tot = E_n*208+E_0
    REL_GAMMA = E_tot/E_0
    REL_BETA = np.sqrt(1 - REL_GAMMA ** (-2))
    pc = E_tot*REL_BETA
    momentum = pc
    ex = exn / (REL_BETA * REL_GAMMA)
    ey = eyn / (REL_BETA * REL_GAMMA)

    Brho = pc * 3.3356/54 # conversions to have Tm
    PARENT_DIR = "."



    mad = Madx(stdout=False)
    ###
    ######
    ##########   LINES TO IMPORT FROM PS official (?) sequence WES
    if corr_seq == False:
        mad.input(requests.get("https://gitlab.cern.ch/acc-models/acc-models-ps/-/raw/2022/ps_mu.seq").text)
        mad.input(requests.get("https://gitlab.cern.ch/acc-models/acc-models-ps/-/raw/2022/ps_ss.seq").text)
        mad.input(requests.get(
            "https://gitlab.cern.ch/acc-models/acc-models-ps/-/raw/2022/scenarios/east/4_slow_extraction/ps_se_east.str").text)
    else:
        mad.call('./clean_seq/corr_gitlab.cern.ch_acc-models_acc-models-ps_-_raw_2022_ps_mu.seq.txt')
        mad.call('./clean_seq/corr_gitlab.cern.ch_acc-models_acc-models-ps_-_raw_2022_ps_ss.seq.txt')
        mad.call('./clean_seq/corr_gitlab.cern.ch_acc-models_acc-models-ps_-_raw_2022_scenarios_east_4_slow_extraction_ps_se_east.str.txt')

    mad.input(f' BEAM, PARTICLE=Pb54, MASS=193.737692, CHARGE=54., ENERGY=193.737692 + {E_n}*208;')
    mad.input(f"BRHO      = BEAM->PC * 3.3356/54;")
    mad.use(sequence="PS")
    #### To test ion parameters
    # mad.input('PC = BEAM->PC;')
    # mad.input('ETOT = BEAM->ENERGY;')
    # mad.input('MASS0 = BEAM->MASS;')
    # print('ion parameters checking')
    # print(f'E_tot = {mad.globals.etot}, PC = {mad.globals.pc}, mass = {mad.globals.mass0} ')
    ############################### turn off 23 for ions BUMPER ELECTROSTATIC
    mad.input("kPEBSW23 = 0;")
    mad.input(
        'SELECT, FLAG=TWISS, COLUMN=NAME,KEYWORD,S,L,K1L,KEYWORD,BETX,ALFX,X,DX,PX,DPX,MUX,BETY,ALFY,Y,DY,PY,DPY,MUY,APERTYPE,APER_1,APER_2,APER_3,APER_4,KMIN,RE11,RE12,RE21,RE22,RE33,RE34,RE43,RE44,RE16,RE26;')
    mad.input('savebeta, label=bumped23, place = PR.BPM23;')


    # Makethin with 4 slices
    QUAD_SLICE = 4
    mad.use(sequence='ps')
    mad.select(flag='makethin', class_='rbend', slice_=QUAD_SLICE)
    mad.select(flag='makethin', class_='quadrupole', slice_=QUAD_SLICE)
    mad.select(flag='makethin', class_='sbend', slice_=QUAD_SLICE)
    mad.select(flag='makethin', class_='hkicker', slice_=QUAD_SLICE, thick=False)
    mad.select(flag='makethin', class_='sextupole', slice_=2)
    mad.makethin(sequence='ps')

    # Match tune using PFWs
    pfw_dict_on_resonance, tune_info_on_resonance = tune_match(mad, Qx=QX_TARGET, Qxp=QPX_TARGET, Qy=QY_TARGET,
                                                               Qyp=QPY_TARGET, verbose=verbose)
    # pfw_dict_on_resonance, tune_info_on_resonance = Tmatch(mad,Qx=QX_TARGET,Qxp=QPX_TARGET,Qy=QY_TARGET,Qyp=QPY_TARGET, p=p, ex=ex, ey=ey)

    mad.input("k1prpfwf = " + str(pfw_dict_on_resonance["k1prpfwf"]) + ";")
    mad.input("k1prpfwd = " + str(pfw_dict_on_resonance["k1prpfwd"]) + ";")
    mad.input("k2prpfwf = " + str(pfw_dict_on_resonance["k2prpfwf"]) + ";")
    mad.input("k2prpfwd = " + str(pfw_dict_on_resonance["k2prpfwd"]) + ";")
    twiss_after_makethin = mad.twiss().dframe()

    print("> Import Complete, MAD-X is ready to go!")

    Qx = tune_info_on_resonance['Qx']
    Qxp = tune_info_on_resonance['Qxp']
    if verbose:
        print(f"\n Qx = {Qx}, Qxp = {Qxp}")

    ############################## Convert Line to Xsuite   #####################
    mad.use(sequence="PS")


    if corr_seq == False:
        line = xt.Line.from_madx_sequence(
            mad.sequence(), allow_thick=True)
    else:
        line = xt.Line.from_madx_sequence(
            mad.sequence(), allow_thick=True, deferred_expressions=True)

    ## WHY SHOULD I CALL THE PARTICLE REF ALSO HERE??

    line.particle_ref = xp.Particles(
        mass0=E_0*1e9,                 # I could change with the output from the energy function!! without accepting it as a value in the funciton
        q0=54,
        beta0=REL_BETA)

    ######################### Calculate Frev ------> REVOLUTION FREQUENCY | WES   ######################
    line.build_tracker(_context=ctx)
    twiss_parameters = line.twiss(method='4d')
    if verbose:
        print(f'CHROMATICITY IN XSUITE :{twiss_parameters.dqx}')  ######
    trev = twiss_parameters['T_rev0']
    frev = 1 / trev
    line.unfreeze()
    if verbose:
        print(f"frev = {frev}")


    # N_turn override
    if n_turns is None or n_turns == 0 or n_turns == '0':
        if verbose:
            print("> n_turns not specified, calculating from time")
            print(f">This sim will run for {n_turns} turns")
        n_turns = int(float(time) * frev)
    else:
        n_turns = int(n_turns)
        if verbose:
            print("> n_turns overridden")

    rf10MHz_enable = True if (rf10MHz_enable == '1') | (rf10MHz_enable == 1) | (rf10MHz_enable == 'True') else False

    ####################  ADDING THE SEPTUM        ############################

    septum = xt.LimitRect(
        min_x=SEPTA_X_MM * 0.001,  # in meters
    )

    line.insert_element(
        element=septum,
        name="SEPTUM",
        index='pe.smh57'
    )

    # TFB Config
    rfko_kick = slwex.kick_angle(float(volt), Brho)
    if norm_gain:
        rfko_kick = rfko_kick / Brho
    ############################# Define your parameters for a single chirp ######################

    start_freq = frev/3 * freq_start_ratio
    end_freq = frev/3 * freq_end_ratio

    # Generate the repeated chirp

    if chirp_type == 'linear':
        t, chirp_signal = ph.generate_chirp(start_freq, end_freq, duration, sampling_freq)
    else:  # non ci sono altre opzioni
        t, chirp_signal = ph.generate_chirp_non_linear(start_freq, end_freq, duration, sampling_freq)

    # Exciter config; the duration parameter tells it to repeat the chirp_signal
    rfko_exciter = xt.Exciter(
        _context=ctx,
        samples=chirp_signal,
        sampling_frequency=sampling_freq,
        frev=frev,
        duration=float(time),
        start_turn=0,
        knl=[rfko_kick]
    )

    line.insert_element(
        element=rfko_exciter,
        name=f'EXCITER',
        index='pr.kfb97'
    )
    if verbose:
        print('exciter Done')

    ########################## RF Cavity Config; voltage 0 if disabled   ########################
    if rf10MHz_enable:
        RF_VOLTAGE = 130e3  # V
    else:
        RF_VOLTAGE = 0

    RF_HARMONIC = 8
    RF_FREQUENCY = frev * RF_HARMONIC

    PHASE = 0  # rising-edge

    rf_cavity = xt.Cavity(
        _context=ctx,
        voltage=RF_VOLTAGE,
        frequency=RF_FREQUENCY,
        lag=PHASE
    )

    line.insert_element(
        element=rf_cavity,
        name="rf_cavity",
        at_s=0
    )
    ######################

    ####################### Insert a monitor to record the particles MONITOR AT THE SEPTUM  #############
    monitor = xt.ParticlesMonitor(
        num_particles=n_part,
        start_at_turn=0,
        stop_at_turn=n_turns,
        auto_to_numpy=True
    )

    line.insert_element(index='pe.smh57', element=monitor, name='septa_monitor')

    line.build_tracker(_context=ctx)
    line.particle_ref = xp.Particles(
        mass0=E_0*1e9,
        q0=54,
        beta0=REL_BETA
    )

    ###############################     PARTICLE GENERARION--- from bunch if RF #####################

    if rf10MHz_enable:
        if verbose:
            print("RF present; generating from generate_matched_gaussian_bunch()")
        particles = xp.generate_matched_gaussian_bunch(
            line=line,
            num_particles=int(n_part),
            nemitt_x=exn,
            nemitt_y=eyn,
            sigma_z=4.3,
            total_intensity_particles=30e10
        )
    else:
        if verbose:
            print("RF not present")
        sigmas, sigmas_p = xp.generate_2D_gaussian(int(n_part))

        try:
            z_array = scipy.stats.uniform.rvs(loc=0,
                                              scale=twiss_parameters['circumference'],
                                              size=int(n_part))
            delta_array = scipy.stats.norm.rvs(loc=0, scale=DPP_FACTOR / 2, size=int(n_part))
            particles = line.build_particles(
                method='4d',
                delta=delta_array,
                zeta=z_array,
                nemitt_x=exn,
                nemitt_y=eyn,
                x_norm=sigmas,
                px_norm=sigmas_p,
                y_norm=sigmas,
                py_norm=sigmas_p,
            )

            # particles.zeta = z_array
            # particles.delta = delta_array
        except AssertionError:
            print("Failed line.build, trying xp.build_particles()")
            z_array = scipy.stats.uniform.rvs(loc=0,
                                              scale=twiss_parameters['circumference'],
                                              size=int(n_part))
            delta_array = scipy.stats.norm.rvs(loc=0, scale=DPP_FACTOR / 2, size=int(n_part))
            particles = xp.build_particles(
                particle_ref=line.particle_ref,
                method='4d',
                delta=delta_array,
                zeta = z_array,
                nemitt_x=exn,
                nemitt_y=eyn,
                x_norm=sigmas,
                px_norm=sigmas_p,
                y_norm=sigmas,
                py_norm=sigmas_p,
            )

            # particles.zeta = z_array
            # particles.delta = delta_array
        else:
            if verbose:
                print("Generating from line.build_particles()")

    # RECOMPUTE THE FINAL TWISS because twiss_parameters is calculated before adding SEPTA-TRANSVERSE_EXCITER
    final_twiss = line.twiss(method='4d')

    if verbose:
        print(
            f'Summary of the used parameters: \n  \n Gain = {volt} \n \n RF_generation = {rf10MHz_enable} \n \n exn,exy = {exn}, {eyn} \n \n DPP = {DPP_FACTOR} \n \n' +
            f' Chirp frequencies = ({start_freq},{end_freq}) \n \n revolution freq = {frev} \n \n X Tune = {Qx} \n \n X Chromaticity = {Qxp} \n \n Septa x lim [mm] = {SEPTA_X_MM}')
    params_out = dict(REL_GAMMA=REL_GAMMA, exn=exn, pc=pc, eyn=eyn, duration=duration, volt=volt, time=time,
                      n_part=n_part,
                      n_turns=n_turns, freq_start_ratio=freq_start_ratio, freq_end_ratio=freq_end_ratio,
                      sampling_freq=sampling_freq,
                      QY_TARGET=QY_TARGET, QPY_TARGET=QPY_TARGET, QX_TARGET=QX_TARGET, QPX_TARGET=QPX_TARGET,
                      rf10MHz_enable=rf10MHz_enable,
                      DPP_FACTOR=DPP_FACTOR, SEPTA_X_MM=SEPTA_X_MM, ctx=ctx, return_form=return_form, REL_BETA=REL_BETA,
                      momentum=momentum, Brho=Brho)

    if return_form == 'dict':
        output = dict(line=line, particles=particles, monitor=monitor, twiss=final_twiss, params=params_out, mad=mad)
        return output










def build_line_baremachine(ions = True,E_nucleons=0.650,dir='mad_file',madout=False,print_info=False,makethin=True):
    '''
    :param ions: True for Ions
    :param print_info: True for printing info of the line and the beam parameters
    :param E_0: rest energy in Gev
    :param E_tot: Total energy in Gev (E_0+T)
    :param out_params: bool, if you want to output a dictionary of parameter
    :param madout: bool, if you want to output the mad object
    :return: Line, depending on option (mad OR params)
    '''
    #,E_0=193.766308,E_tot =328.966308
    rel_params = ph.energy(ions,E_nucleons)
    E_0 = rel_params['E_0']
    if ions:
        particle = 'Pb54'
        charge = 54
    else:
        charge = 1
        particle = 'PROTON'


    mad = Madx(stdout=False)




    ## MAIN UNITS
    mad.call(f'./{dir}/final_corr_STANDARD_PR_YETS 2022-2023_20-MAR-2023.seq')

    mad.input(f'BEAM, PARTICLE={particle}, MASS={rel_params["E_0"]}, CHARGE={charge}., ENERGY={rel_params["E_tot"]};')
    mad.input(f'BRHO = BEAM->PC * 3.3356 /{charge}.;') # Beam rigidity
    mad.input('PC=BEAM->PC;') # more for debugging than else
    mad.call(f'./{dir}/ps_ion_bare_machine.str')


    ########## MAKETHIN
    if makethin:
        QUAD_SLICE = 4
        mad.use(sequence='ps')
        mad.select(flag='makethin', class_='rbend', slice_=QUAD_SLICE)
        mad.select(flag='makethin', class_='quadrupole', slice_=QUAD_SLICE)
        mad.select(flag='makethin', class_='sbend', slice_=QUAD_SLICE)
        mad.select(flag='makethin', class_='hkicker', slice_=QUAD_SLICE, thick=False)
        mad.select(flag='makethin', class_='sextupole', slice_=2)
        mad.makethin(sequence='ps')


    ### Twiss to GET SOME PARAMETER CALCULATED IN MADX
    mad.input('SELECT, FLAG=TWISS, CLEAR;')
    mad.input('SELECT, FLAG=TWISS, FULL;')
    mad.twiss()
    ## Found in Xsuite --> momentum_compaction_factor
    mad.input('DQX = TABLE(SUMM,DQ1);')
    mad.input('DQY = TABLE(SUMM,DQ2);')
    mad.input('QX = TABLE(SUMM,Q1);')
    mad.input('QY = TABLE(SUMM,Q2);')
    mad.input('GAMMATR = TABLE(SUMM,GAMMATR);')

    if print_info:
         print(f'the value for PC in madx is {mad.globals.pc} Gev')
         print(f'the chromaticities in madx are : dqx = {mad.globals.dqx},   dqy = {mad.globals.dqy}')
         print(f'gamma transition is : {mad.globals.gammatr}')
         #print(mad.globals.gammatr)
         print(f'Brho of mad is {mad.globals.brho}')
         #print(mad.globals.brho)


    # Relativistic params calcualted by energy
    gamma = rel_params['gamma']
    beta = rel_params['beta']
    pc = rel_params['pc']

    print('--------------------------ENERGY -------------')
    print(f'gamma = {gamma}   beta = {beta}     pc = {pc}')
    print('----------------------------------------------------------')
    ## LINE TO XSUITE
    ctx = xo.ContextCpu()
    line = xt.Line.from_madx_sequence(mad.sequence(),allow_thick=True,deferred_expressions=True)
    line.particle_ref = xp.Particles(
        mass0=E_0*1e9,
        q0=charge,
        beta0=beta)

    line.build_tracker(_context=ctx)

    if print_info:
        twiss = line.twiss(method="4d")
        print(f'chromaticities in Xsuite are : dqx = {twiss.dqx}, dqy = {twiss.dqy}')

    ########## OUTPUT
    if madout :
        print('returning a dictionary with the mad object and the line and the params')
        out = dict(params=rel_params,mad=mad,line=line)
        return out
    else:
        return dict(line=line,params=rel_params)




def henon_map_line(ions=True,E_nucleons=200,alfx=None,betx=None,alfy=None,bety=None,qx=6.32000,
                   qy=6.2445267,dqx=4.068617,dqy=-10.248037,dx=2.985807,k2nl=-2.867035,S=None):
    '''TODO 1) check the correct environment, 2) add other custom parameters to input for henon map'''
    rel_params = ph.energy(ions, E_nucleons)
    E_0 = rel_params['E_0']
    if ions:
        particle = 'Pb54'
        charge = 54
    else:
        charge = 1
        particle = 'PROTON'
    # this is the order accepted for the henon map
    twiss_params = [alfx, betx, alfy, bety]
    # Default value from the PS line
    std_twiss = [0.838476296628905,13.784753228063757,
     -0.7032505108349398,21.134855356093915]
    for i,x in enumerate(twiss_params):
        if x is None:
            twiss_params[i] = std_twiss[i]

    if S is not None:
        # Overrides the values of k2nl
        k2nl = 2*S/(twiss_params[1]**(3/2))
    else:
        S = k2nl*twiss_params[1]**(3/2)/2

    gamma = rel_params['gamma']
    beta = rel_params['beta']
    pc = rel_params['pc']

    omega_x = 2*np.pi*qx
    omega_y = 2*np.pi*qy
    Henon_ele = xt.Henonmap(omega_x=omega_x,omega_y=omega_y,dqx=dqx,dqy=dqy,dx=dx,
                            twiss_params=twiss_params,norm=False,multipole_coeffs=[k2nl])
    line = xt.Line(elements=[Henon_ele],element_names=['Henon_sextupole'])

    line.particle_ref = xp.Particles(
        mass0=E_0*1e9,
        q0=charge,
        beta0=beta)
    henon_params = dict(qx=qx,qy=qy,dqx=dqx,dqy=dqy,dx=dx,
                            twiss_params=twiss_params,norm=False,multipole_coeffs=[k2nl],S=S)

    return dict(line=line,params = rel_params,henon_params=henon_params)


def rfko_setup_old(REL_GAMMA=25.598474067, exn=2e-6, p=5.392, eyn=2e-6, duration=1 / 1000, volt=0, time=0.004,
                   n_part=1000,
                   n_turns=1000, freq_start_ratio=90 / 100, freq_end_ratio=110 / 100, sampling_freq=1e9,
                   QY_TARGET=0.256, QPY_TARGET=-0.242, QX_TARGET=0.325, QPX_TARGET=-0.5, rf10MHz_enable='0',
                   DPP_FACTOR=0, SEPTA_X_MM=-60, ctx=xo.ContextCpu(), chirp_type='linear', verbose=True,
                   norm_gain=False,dir=None, corr_seq=True, return_form='dict'):
    # Beam parameters

    every = ['freq_start_ratio', 'freq_end_ratio', 'duration', 'sampling_freq', 'REL_BETA', 'REL_GAMMA', 'exn',
             'eyn', 'p', 'Brho', 'QX_TARGET', 'QPX_TARGET', 'QY_TARGET', 'QPY_TARGET', 'volt', 'rf10MHz_enable', 'time',
             'n_part', 'n_turns', 'SEPTA_X_MM']

    REL_BETA = np.sqrt(1 - REL_GAMMA ** (-2))
    momentum = p
    ex = exn / (REL_BETA * REL_GAMMA)
    ey = eyn / (REL_BETA * REL_GAMMA)
    Brho = p * 3.3356
    PARENT_DIR = "."
    os.makedirs(PARENT_DIR, exist_ok=True)
    if verbose:
        print(f"Parent Directory: {PARENT_DIR}")
    mad = Madx(stdout=False)

    ##########
    ##########   LINES TO IMPORT FROM PS official (?) sequence WES
    if dir is None:
        dir = 'clean_seq'
    if corr_seq == False:
        mad.input(requests.get("https://gitlab.cern.ch/acc-models/acc-models-ps/-/raw/2022/ps_mu.seq").text)
        mad.input(requests.get("https://gitlab.cern.ch/acc-models/acc-models-ps/-/raw/2022/ps_ss.seq").text)
        mad.input(requests.get(
            "https://gitlab.cern.ch/acc-models/acc-models-ps/-/raw/2022/scenarios/east/4_slow_extraction/ps_se_east.str").text)
    else:
        mad.call(f'./{dir}/corr_gitlab.cern.ch_acc-models_acc-models-ps_-_raw_2022_ps_mu.seq.txt')
        mad.call(f'./{dir}/corr_gitlab.cern.ch_acc-models_acc-models-ps_-_raw_2022_ps_ss.seq.txt')
        mad.call(
            f'./{dir}/corr_gitlab.cern.ch_acc-models_acc-models-ps_-_raw_2022_scenarios_east_4_slow_extraction_ps_se_east.str.txt')

    mad.command.beam(
        particle="PROTON",
        pc=p,
        ex=ex,
        ey=ey,
        charge=1
    )
    mad.input(f"BRHO      = BEAM->PC * 3.3356;")
    mad.use(sequence="PS")

    ############################### turn off 23 for ions BUMPER ELECTROSTATIC
    mad.input("kPEBSW23 = 0;")
    mad.input(
        'SELECT, FLAG=TWISS, COLUMN=NAME,KEYWORD,S,L,K1L,KEYWORD,BETX,ALFX,X,DX,PX,DPX,MUX,BETY,ALFY,Y,DY,PY,DPY,MUY,APERTYPE,APER_1,APER_2,APER_3,APER_4,KMIN,RE11,RE12,RE21,RE22,RE33,RE34,RE43,RE44,RE16,RE26;')
    mad.input('savebeta, label=bumped23, place = PR.BPM23;')

    # Makethin with 4 slices
    QUAD_SLICE = 4
    mad.use(sequence='ps')
    mad.select(flag='makethin', class_='rbend', slice_=QUAD_SLICE)
    mad.select(flag='makethin', class_='quadrupole', slice_=QUAD_SLICE)
    mad.select(flag='makethin', class_='sbend', slice_=QUAD_SLICE)
    mad.select(flag='makethin', class_='hkicker', slice_=QUAD_SLICE, thick=False)
    mad.select(flag='makethin', class_='sextupole', slice_=2)
    mad.makethin(sequence='ps')

    # Match tune using PFWs
    pfw_dict_on_resonance, tune_info_on_resonance = tune_match(mad, Qx=QX_TARGET, Qxp=QPX_TARGET, Qy=QY_TARGET,
                                                               Qyp=QPY_TARGET, verbose=verbose)
    # pfw_dict_on_resonance, tune_info_on_resonance = Tmatch(mad,Qx=QX_TARGET,Qxp=QPX_TARGET,Qy=QY_TARGET,Qyp=QPY_TARGET, p=p, ex=ex, ey=ey)

    mad.input("k1prpfwf = " + str(pfw_dict_on_resonance["k1prpfwf"]) + ";")
    mad.input("k1prpfwd = " + str(pfw_dict_on_resonance["k1prpfwd"]) + ";")
    mad.input("k2prpfwf = " + str(pfw_dict_on_resonance["k2prpfwf"]) + ";")
    mad.input("k2prpfwd = " + str(pfw_dict_on_resonance["k2prpfwd"]) + ";")
    twiss_after_makethin = mad.twiss().dframe()

    print("> Import Complete, MAD-X is ready to go!")

    Qx = tune_info_on_resonance['Qx']
    Qxp = tune_info_on_resonance['Qxp']  # Questa dovrebbe esssere la cromaticità
    if verbose:
        print(f"\n Qx = {Qx}, Qxp = {Qxp}")

    ############################## Convert Line to Xsuite   #####################
    mad.use(sequence="PS")

    if corr_seq == False:
        line = xt.Line.from_madx_sequence(
            mad.sequence(), allow_thick=True)

    else:
        line = xt.Line.from_madx_sequence(
            mad.sequence(), allow_thick=True, deferred_expressions=True)

    line.particle_ref = xp.Particles(
        mass0=xp.PROTON_MASS_EV,
        q0=1,
        p0c=p * scipy.constants.c)

    ######################### Calculate Frev ------> REVOLUTION FREQUENCY | WES   ######################
    line.build_tracker(_context=ctx)
    twiss_parameters = line.twiss(method='4d')
    if verbose:
        print(f'CHROMATICITY IN XSUITE :{twiss_parameters.dqx}')  ######
    trev = twiss_parameters['T_rev0']
    frev = 1 / trev
    line.unfreeze()
    if verbose:
        print(f"frev = {frev}")
    # frev = 412657.22413208015

    # N_turn override
    if n_turns is None or n_turns == 0 or n_turns == '0':
        if verbose:
            print("> n_turns not specified, calculating from time")
            print(f">This sim will run for {n_turns} turns")
        n_turns = int(float(time) * frev)
    else:
        n_turns = int(n_turns)
        if verbose:
            print("> n_turns overridden")

    rf10MHz_enable = True if (rf10MHz_enable == '1') | (rf10MHz_enable == 1) | (rf10MHz_enable == 'True') else False

    ####################  ADDING THE SEPTUM        ############################

    septum = xt.LimitRect(
        min_x=SEPTA_X_MM * 0.001,  # in meters
    )

    line.insert_element(
        element=septum,
        name="SEPTUM",
        index='pe.smh57'
    )

    # TFB Config
    rfko_kick = slwex.kick_angle(float(volt), Brho)
    if norm_gain:
        rfko_kick = rfko_kick / Brho
    ############################# Define your parameters for a single chirp ######################

    start_freq = frev * freq_start_ratio
    end_freq = frev * freq_end_ratio

    # Generate the repeated chirp

    if chirp_type == 'linear':
        t, chirp_signal = ph.generate_chirp(start_freq, end_freq, duration, sampling_freq)
    else:  # non ci sono altre opzioni
        t, chirp_signal = ph.generate_chirp_non_linear(start_freq, end_freq, duration, sampling_freq)

    # Exciter config; the duration parameter tells it to repeat the chirp_signal
    rfko_exciter = xt.Exciter(
        _context=ctx,
        samples=chirp_signal,
        sampling_frequency=sampling_freq,
        frev=frev,
        duration=float(time),
        start_turn=0,
        knl=[rfko_kick]
    )

    line.insert_element(
        element=rfko_exciter,
        name=f'EXCITER',
        index='pr.kfb97'
    )
    if verbose:
        print('exciter Done')

    ########################## RF Cavity Config; voltage 0 if disabled   ########################
    if rf10MHz_enable:
        RF_VOLTAGE = 130e3  # V
    else:
        RF_VOLTAGE = 0

    RF_HARMONIC = 8
    RF_FREQUENCY = frev * RF_HARMONIC

    PHASE = 0  # rising-edge

    rf_cavity = xt.Cavity(
        _context=ctx,
        voltage=RF_VOLTAGE,
        frequency=RF_FREQUENCY,
        lag=PHASE
    )

    line.insert_element(
        element=rf_cavity,
        name="rf_cavity",
        at_s=0
    )
    ######################

    ####################### Insert a monitor to record the particles MONITOR AT THE SEPTUM  #############
    monitor = xt.ParticlesMonitor(
        num_particles=n_part,
        start_at_turn=0,
        stop_at_turn=n_turns,
        auto_to_numpy=True
    )

    line.insert_element(index='pe.smh57', element=monitor, name='septa_monitor')

    line.build_tracker(_context=ctx)
    line.particle_ref = xp.Particles(
        mass0=xp.PROTON_MASS_EV,
        q0=1,
        p0c=p * scipy.constants.c
    )

    ###############################     PARTICLE GENERARION--- from bunch if RF #####################

    if rf10MHz_enable:
        if verbose:
            print("RF present; generating from generate_matched_gaussian_bunch()")
        particles = xp.generate_matched_gaussian_bunch(
            line=line,
            num_particles=int(n_part),
            nemitt_x=exn,
            nemitt_y=eyn,
            sigma_z=4.3,
            total_intensity_particles=30e10
        )
    else:
        if verbose:
            print("RF not present")
        sigmas, sigmas_p = xp.generate_2D_gaussian(int(n_part))

        try:
            particles = line.build_particles(
                method='4d',
                delta=DPP_FACTOR,
                nemitt_x=exn,
                nemitt_y=eyn,
                x_norm=sigmas,
                px_norm=sigmas_p,
                y_norm=sigmas,
                py_norm=sigmas_p,
            )
            z_array = scipy.stats.uniform.rvs(loc=0 - twiss_parameters['circumference'] / 2,
                                              scale=twiss_parameters['circumference'],
                                              size=int(n_part))
            delta_array = scipy.stats.norm.rvs(loc=0, scale=DPP_FACTOR / 2, size=int(n_part))
            particles.zeta = z_array
            particles.delta = delta_array
        except AssertionError:
            print("Failed line.build, trying xp.build_particles()")
            particles = xp.build_particles(
                particle_ref=line.particle_ref,
                method='4d',
                delta=DPP_FACTOR,
                nemitt_x=exn,
                nemitt_y=eyn,
                x_norm=sigmas,
                px_norm=sigmas_p,
                y_norm=sigmas,
                py_norm=sigmas_p,
            )
            z_array = scipy.stats.uniform.rvs(loc=0 - twiss_parameters['circumference'] / 2,
                                              scale=twiss_parameters['circumference'],
                                              size=int(n_part))
            delta_array = scipy.stats.norm.rvs(loc=0, scale=DPP_FACTOR / 2, size=int(n_part))
            particles.zeta = z_array
            particles.delta = delta_array
        else:
            if verbose:
                print("Generating from line.build_particles()")

    # RECOMPUTE THE FINAL TWISS because twiss_parameters is calculated before adding SEPTA-TRANSVERSE_EXCITER
    final_twiss = line.twiss(method='4d')

    if verbose:
        print(
            f'Summary of the used parameters: \n  \n Gain = {volt} \n \n RF_generation = {rf10MHz_enable} \n \n exn,exy = {exn}, {eyn} \n \n DPP = {DPP_FACTOR} \n \n' +
            f' Chirp frequencies = ({start_freq},{end_freq}) \n \n revolution freq = {frev} \n \n X Tune = {Qx} \n \n X Chromaticity = {Qxp} \n \n Septa x lim [mm] = {SEPTA_X_MM}')
    params_out = dict(REL_GAMMA=REL_GAMMA, exn=exn, p=p, eyn=eyn, duration=duration, volt=volt, time=time,
                      n_part=n_part,
                      n_turns=n_turns, freq_start_ratio=freq_start_ratio, freq_end_ratio=freq_end_ratio,
                      sampling_freq=sampling_freq,
                      QY_TARGET=QY_TARGET, QPY_TARGET=QPY_TARGET, QX_TARGET=QX_TARGET, QPX_TARGET=QPX_TARGET,
                      rf10MHz_enable=rf10MHz_enable,
                      DPP_FACTOR=DPP_FACTOR, SEPTA_X_MM=SEPTA_X_MM, ctx=ctx, return_form=return_form, REL_BETA=REL_BETA,
                      momentum=momentum, Brho=Brho)

    if return_form == 'dict':
        output = dict(line=line, particles=particles, monitor=monitor, twiss=final_twiss, params=params_out, mad=mad)
        return output





from setuptools import setup, find_packages

with open("README.md",
          "r", encoding="utf-8") as file:
    long_description = file.read()

### xcoll not considered for now

setup(
    name='RFKO_Xsuite',
    version='0.1.0',
    packages=find_packages(exclude=['tests*']),
    license='MIT',
    description='A package for wrapping up some funcitonalities from Xsuite',
    long_description=long_description,
    install_requires=['xsuite','xpart','xobjects','xtrack','xcoll','cpymad','matplotlib'
        ,'numpy', 'pandas'],  # add any additional packages you want to include here
    url='http://github.com/username/Package_',  # replace with your package's url
    author='Wesley',
    author_email='wesley.scarpa@cern.ch'
)







import numpy as np
import scipy
import matplotlib.pyplot as plt
import inspect

import xtrack as xt
import xpart as xp
import xobjects as xo

from . import xsuite_helper as xh
from . import slowex_helper as slwex
from . import phy_helper as ph






def build_particles2(line, n_part=1000, DPP_FACTOR=2.35e-3, debunched_co=True,ramping_sex=False,exn=2.64378704e-06, eyn= 4.7590473e-07, twiss=None, out_params=False):
    ''' TODO
        - add bunched generation options

    :param line:
    :param n_part:
    :param DPP_FACTOR:
    :param debunched_co:
    :param exn:
    :param eyn:
    :param twiss:
    :param out_params:
    :return:
    '''
    if out_params: # can be useful if not used with the wrapper
        params = locals()

    if ramping_sex:
        ##  account for the real setup of the first turns
        xse_names = ['pr.xse01.a', 'pr.xse01.b', 'pr.xse07']
        setter = xt.MultiSetter(line, xse_names, field='k2')  # multisetter obj
        xse0 = setter.get_values()  # initial/nominal values
        setter.set_values([0., 0., 0.])  # changes of value

    if twiss is None:
        twiss = line.twiss(method='4d')

    twiss_df = twiss.to_pandas()
    sigmas, sigmas_p = xp.generate_2D_gaussian(int(n_part))

    z_array = scipy.stats.uniform.rvs(loc=0,
                                      scale=twiss['circumference'],
                                      size=int(n_part))

    # momentum off set will be gauss distributed along zero and DPP_FACTOR will be the standard deviation
    delta_array = scipy.stats.norm.rvs(loc=0, scale=DPP_FACTOR , size=int(n_part))
    if debunched_co:
        # Transport the particles along the line adding the orbit distortion to each point
        sigmas,sigmas_p = xh.nearest_betx(twiss_df,sigmas,sigmas_p,z_array)


    particles = line.build_particles(
        method='4d',
        zeta = z_array,
        delta=delta_array,
        nemitt_x=exn,
        nemitt_y=eyn,
        x_norm=sigmas,
        px_norm=sigmas_p,
        y_norm=sigmas,
        py_norm=sigmas_p,
    )
    # particles.zeta = z_array
    # particles.delta = delta_array

    # RESET THE LINE TO THE NOMINAL VALUES
    if ramping_sex:
        setter.set_values(xse0)
    if out_params:
        return particles, params
    else:
        return particles


def insert_septa_monitor(line, SEPTA_X_MM=-75, n_turns=1000, n_part=1000, out_monitor=True,out_params=False):
    ''' It is not general in where we could place the monitor, mostly beacause for my purposes I don't need this generalization
        and the same is true for every PS sequences. Eventually this will be generalized to different sequences
    '''
    line.unfreeze()
    if out_params:
        params = locals()
    ################################## SEPTUM
    septum = xt.LimitRect(
        min_x=SEPTA_X_MM * 0.001
    )

    line.insert_element(
        element=septum,
        name="SEPTUM",
        index='pe.smh57'
    )
    ##### end SEPTUM

    #######################  MONITOR AT THE SEPTUM
    monitor = xt.ParticlesMonitor(
        num_particles=n_part,
        start_at_turn=0,
        stop_at_turn=n_turns,
        auto_to_numpy=True
    )

    line.insert_element(index='pe.smh57', element=monitor, name='septa_monitor')

    ####### end MONITOR
    if out_monitor&out_params:
        return monitor,params
    elif out_monitor:
        return monitor


def plot_stable(out_dict, s=2, figsize=(12, 7), interpolate=False, wrongf=False, center=True):
    twiss_df = out_dict['twiss'].to_pandas()
    monitor = out_dict['monitor']
    S_strength, S_mu = slwex.Virtual_sextupole_complete(out_dict)  # compute sextupole strength and mu
    dmu = twiss_df[twiss_df.name == 'pe.smh57'].mux.iloc[0] * 2 * np.pi - S_mu
    # dmu from the virtual sextupole to know the rotation

    Monitor = xh.Normalize_monitor(out_dict, normalize_particles=False, center=center)
    color = np.arange(out_dict['params']['n_turns'])
    colors = np.tile(color, (out_dict['params']['n_part'], 1))

    extracted_mask = Monitor.x != 0
    argmin = np.argmin(np.abs(twiss_df.mux * 2 * np.pi - S_mu))
    # INTERPOLATEE BETX AND X
    if (interpolate) & (wrongf is False):
        xi, idxi = np.unique(twiss_df.mux, return_index=True)
        yi = twiss_df.iloc[idxi].betx
        yix = twiss_df.iloc[idxi].x
        betxinterp = np.interp(S_mu / (2 * np.pi), xi, yi)
        xinterp = np.interp(S_mu / (2 * np.pi), xi, yix)

        print(f'X: nearest={twiss_df.iloc[argmin]["x"]}, interpolated={xinterp}')
        print(f'BETX: nearest= {twiss_df.iloc[argmin]["betx"]}, interpolated= {betxinterp}')
        # INTERPOLATED
        Xo = xinterp / np.sqrt(betxinterp)

    elif wrongf is False:
        ### NEAREST method
        Xo = twiss_df.iloc[argmin]['x'] / np.sqrt(twiss_df.iloc[argmin]['betx'])
    else:
        # WRONG version THAT WORKs BETTER
        Xo = twiss_df.iloc[argmin]['x'] / twiss_df.iloc[argmin]['betx']

        ## CLOSED ORBIT CAUSES A CHANGE IN TUNE SHIFT
    dq0 = (out_dict['twiss'].qx - 6 - 1 / 3)  # normal as without distorsions and dispersion
    dq = dq0 - S_strength * Xo / (2 * np.pi)  # distortion correction

    ####check if DPP_FACTOR NON ZERO ADD CHROMATICITY CORRECTION
    # print(f'The tune is {out_dict["twiss"].qx}')
    # print(f'The tune distance is {dq} and the Sextupole strength {S_strength}')
    # print(f'their ratio (proportional to the area) {dq/S_strength}')

    plt.figure(figsize=figsize)
    plt.scatter(Monitor.x[extracted_mask], Monitor.px[extracted_mask], s=s, c=colors[extracted_mask])
    plt.colorbar(label='Turns')
    if center is False:
        # TRANLATION OF THE TRIANGLE AT THE MONITOR CLOSED ORBIT
        xt = twiss_df[twiss_df.name == 'pe.smh57'].x.iloc[0]
        pxt = twiss_df[twiss_df.name == 'pe.smh57'].px.iloc[0]
        betx = twiss_df[twiss_df.name == 'pe.smh57'].betx.iloc[0]
        alfx = twiss_df[twiss_df.name == 'pe.smh57'].alfx.iloc[0]
        twiss_df[twiss_df.name == 'pe.smh57'].alfx.iloc[0]
        Xt = xt / np.sqrt(betx)
        Pxt = Xt * alfx + pxt * np.sqrt(betx)
        t_vector = np.array([Xt, Pxt])
        slwex.stable_tri_plot(dq, - S_strength, dmu=dmu, linewidth=1, translation_vector=t_vector)

    else:
        slwex.stable_tri_plot(dq, - S_strength, dmu=dmu, linewidth=1)
    plt.title(
        f'Tune =  {round(out_dict["twiss"].qx, 5)}, ratio tune distance to sextupole = {round(dq / S_strength, 5)}')


def insert_rfko(line,Brho=None,pc=None,beta=None,charge=None,volt=0,third_frev=True,twiss=None,chirp_type='linear',
                frev=None,freq_start_ratio=90/100,freq_end_ratio=110/100,sampling_freq=1e9,
                duration=1/2000,time=0.04,ctx = xo.ContextCpu(),out_params=False,kick_function=1):
    if out_params: # useful if used without the wrapper
        params = locals()

    if (Brho is None)|(pc is None)|(beta is None):
        print('no dynamical parameters given, used standard lead ions at 650 Mev per nucleon')
        rel_params = ph.energy()
        pc = rel_params['pc']
        beta = rel_params['beta']
        Brho = rel_params['Brho']
        charge = rel_params['charge']

    if kick_function==1:
        rfko_kick = slwex.kick_angle(float(volt),Brho,pc,beta,charge)
    else:
        rfko_kick = slwex.kick_angle_(float(volt),Brho,pc,beta,charge)

    if twiss is None:
        twiss = line.twiss(method='4d')
    
    # revolution freq
    if frev is None:
        trev = twiss['T_rev0']
        frev = 1 / trev
        print(f'revolution frquency = {frev}')
    # frequancies scanned
    if third_frev:
        start_freq = frev/3 * freq_start_ratio
        end_freq = frev/3 * freq_end_ratio
    else:
        start_freq = frev * freq_start_ratio
        end_freq = frev * freq_end_ratio

    
    if chirp_type =='linear':
        t,chirp_signal = ph.generate_chirp(start_freq, end_freq, duration, sampling_freq)
    else:
        t, chirp_signal = ph.generate_chirp_non_linear(start_freq, end_freq, duration, sampling_freq)
    
    line.unfreeze()
    
    # Exciter config; the duration parameter tells it to repeat the chirp_signal
    rfko_exciter = xt.Exciter(
        _context=ctx,
        samples=chirp_signal,
        sampling_frequency=sampling_freq,
        frev=frev,
        duration=float(time),
        start_turn=0,
        knl=[rfko_kick]
    )
    
    line.insert_element(
        element=rfko_exciter,
        name='EXCITER',
        index='pr.kfb97'
    )
    if out_params:
        params['frev'] = frev
        return params

    
def builder_wrapper(line_dict,twiss=None,third_frev=True,extraction = True,targetx = None,debunched_co=True,n_part=1000,
                    n_turns=None,DPP_FACTOR=2.35e-3, exn=2.64378704e-06, eyn= 4.7590473e-07,
                    SEPTA_X_MM=-75,freq_start_ratio=90/100,freq_end_ratio=110/100,volt=0,sampling_freq=1e9,
                    duration=1/2000,time=None,chirp_type='linear',emittance_norm=True,kick_function=1):
    ''' TODO
    - Better adapt the mechanism to pass the parameters to the functions without passing through the maual creation of the params dict


    :return:
    '''

    line = line_dict['line']
    if twiss is None:
        twiss = line.twiss(method='4d')

    outdict1 = dict(line=line, twiss=twiss)

    ############################################################# MATCHING XTUNE
    matching(outdict1,targetx=targetx,extraction=extraction)
    twiss = outdict1['twiss']

    gamma = line_dict['params']['gamma']
    beta = line_dict['params']['beta']
    pc = line_dict['params']['pc']
    Brho = line_dict['params']['Brho']
    charge = line_dict['params']['charge']

    if emittance_norm: 
        ex = exn/(gamma*beta)
        ey = eyn/(gamma*beta)
    else:
        ex = exn.copy()
        ey = eyn.copy()
        exn = ex*gamma*beta
        eyn = ey*gamma*beta
        
        
    ### SIMULATIONS TIME AND NUMBER OF TURNS
    trev = twiss['T_rev0']
    frev = 1 / trev
    print(f'revolution frequency is {frev}')
    if (time is None) & (n_turns is not None):
        time = trev*n_turns
    if (time is not None) & (n_turns is None):
        n_turns = time/trev
    if (time is None)&(n_turns is None):
        # Turns are the default if None is given
        n_turns=1000
        time =  trev*n_turns

   # PARAMS TO PASS TO THE FUNCTIONS
    params = dict(n_part=n_part,n_turns=n_turns,third_frev=third_frev,pc=pc,Brho=Brho,beta=beta,charge=charge,debunched_co=debunched_co,DPP_FACTOR=DPP_FACTOR, exn=exn, eyn=eyn,
                  SEPTA_X_MM=SEPTA_X_MM,freq_start_ratio=freq_start_ratio,volt=volt,freq_end_ratio=freq_end_ratio,
                  sampling_freq=sampling_freq,duration=duration,gamma=gamma,time=time,ex=ex,ey=ey,
                  chirp_type=chirp_type,emittance_norm=emittance_norm,frev=frev,kick_function=kick_function)
    
    ########
    #This is a check to see if I put al the parameters i wanted in the params dictionary
    # The expected exceptions are : [trev,line,twiss,params]
    #######################################
    # for i in locals().keys():
    #     if i not in params.keys():
    #        print(i)
    # ############
    
    



    ######## AUTOMATIC CHOOSING THE PARAMS FROM PARAMS DICT, THAT ARE USED BY THE FUNCTIONS
    # inspect.getfullargspec
    bpart_args = inspect.getfullargspec(build_particles2).args[1:] # the first is the line non keyword arg
    monito_args = inspect.getfullargspec(insert_septa_monitor).args[1:]
    rfkoel_args = inspect.getfullargspec(insert_rfko).args[1:]
    subpart = dict()
    submon = dict()
    subrko = dict()
    for key,val in params.items():
        if key in bpart_args:
            subpart[key]=val
        if key in monito_args:
            submon[key]=val
        if key in rfkoel_args:
            subrko[key]=val

    #### 1.MONITOR AND SEPTA
    monitor = insert_septa_monitor(line,**submon)
    # 2.RFKO EXCITER
    insert_rfko(line,twiss=twiss,**subrko)
    ### Having unfreezed the line to insert the element, we lose the tracker
    line.build_tracker(_context=xo.ContextCpu())
    # 3.BUILD PARTICLES
    particles = build_particles2(line, twiss=twiss, **subpart)

    ## OUTPUT
    outdict = dict(line=line,twiss=twiss,monitor=monitor,particles=particles,params=params)
    
    return outdict


def matching(outdict, targetx=None,extraction=True):
    ''' TODO
    - More flexible and change the default case for the setting of kf kd
    - add the possibility to match the chromaticity
    '''
    line = outdict['line']
    knobs = ['k1prbhf', 'k1prbhd', 'k2prmp', 'k2prmpj']

    # Value matched to the measurements TUNES OF BARE MACHINE
    vals = [0.05714293659989464, -0.05719922125376693, 0.010731455338789136, -0.019018243990104253]
    for i in range(len(knobs)):
        line.vars[knobs[i]] = vals[i]

    if extraction:
        ####### EXTRACTION UNITS
        line.vars['kpebsw23'] = 0  # from LSA not --> -0.001570859932
        line.vars['kpebsw57'] = 5.544860897340004E-4  # from LSA
        line.vars['kprqse'] = 0.09088224  # Value from LSA try length
        line.vars['kprxse'] = 1.4193310000000103  # value at the steady from LSA


        if targetx is None:  ### STANDARD QXTARGET {6.32 }

            line.vars['kf'] =  0.002252965399686161
            line.vars['kd'] = -0.009766502523392236
            twiss = line.twiss(method='4d')
            outdict['twiss'] = twiss
            print(f'standard value Qx = {twiss.qx}')
        else:
            print(f'the target qx is {targetx}')
            line.match(method='4d', n_steps_max=30, vary=[xt.Vary('kf', step=1e-9), xt.Vary('kd', step=1e-9)],
                       targets=[xt.Target('qx', targetx, tol=0.000005), xt.Target('qy', outdict['twiss'].qy, tol=0.005)])
            twiss = line.twiss(method='4d')
            print(f'after match {twiss.qx}')
            outdict['twiss'] = twiss
    else:
        twiss = line.twiss(method='4d')
        outdict['twiss'] = twiss







import numpy as np
import matplotlib.pyplot as plt
import xobjects as xo
import xtrack as xt
from matplotlib.colors import LinearSegmentedColormap

from . import xsuite_helper as xh
from . import slowex_helper as slwex

from . import phy_helper as ph






#
#
# def insert_rfko_custom(line, signal=None, Brho=None, pc=None, beta=None, tune_factor=1 / 3, volt=0,twiss=None, chirp_type='linear', monofreq=False,
#                        frev=None, freq_start_ratio=90 / 100, freq_end_ratio=110 / 100, sampling_freq=1e9,
#                        duration=1 / 2000, time=None, n_turns=None, ctx=xo.ContextCpu(), out_params=False):
#     ''' it can accept a signal already made up to be repeated
#     '''
#     if out_params:  # useful if used without the wrapper
#         params = locals()
#
#     if (Brho is None) | (pc is None) | (beta is None):
#         print('--> no dynamical parameters given, used standard lead ions at 650 Mev per nucleon')
#         rel_params = ph.energy()
#         pc = rel_params['pc']
#         beta = rel_params['beta']
#         Brho = rel_params['Brho']
#
#     rfko_kick = slwex.kick_angle(float(volt), Brho, pc, beta)
#     if twiss is None:
#         twiss = line.twiss(method='4d')
#
#     # revolution freq
#     if frev is None:
#         trev = twiss['T_rev0']
#         frev = 1 / trev
#         print(f'--> revolution frquency = {frev}')
#
#     # frequancies scanned
#
#     start_freq = frev * tune_factor * freq_start_ratio
#     end_freq = frev * tune_factor * freq_end_ratio
#
#
#     line.unfreeze()
#
#     if signal is not None:
#         plt.title('custom signal fed to exciter')
#         plt.plot(np.arange(len(signal)) / sampling_freq, signal)
#
#     if signal is None:
#         if monofreq:
#             if time is None: # I could delete this simply
#                 time = (n_turns + 3) * twiss.T_rev0
#             t = np.linspace(0,time,int(sampling_freq*time))
#             chirp_signal = np.cos(2*np.pi*frev*tune_factor*t)
#         else:
#             t, chirp_signal = ph.generate_chirp(start_freq, end_freq, duration, sampling_freq)
#         if (time is None) & (n_turns is None):
#             print(
#                 '----- There is no way to infer the total simulation time and for the chirp signal repetion is needed ----')
#
#         elif (time is None) & (n_turns is not None):
#             time = (n_turns + 3) * twiss.T_rev0
#
#         # Exciter config; the duration parameter tells it to repeat the chirp_signal
#         rfko_exciter = xt.Exciter(
#             _context=ctx,
#             samples=chirp_signal,
#             sampling_frequency=sampling_freq,
#             frev=frev,
#             duration=float(time),
#             start_turn=0,
#             knl=[rfko_kick]
#         )
#
#     elif time is None:
#         # Signal already provided with a length matching the total simulation
#         rfko_exciter = xt.Exciter(
#             _context=ctx,
#             samples=signal,
#             sampling_frequency=sampling_freq,
#             frev=frev,
#             start_turn=0,
#             knl=[rfko_kick]
#         )
#     else:
#         rfko_exciter = xt.Exciter(
#             _context=ctx,
#             samples=signal,
#             sampling_frequency=sampling_freq,
#             frev=frev,
#             duration=float(time),
#             start_turn=0,
#             knl=[rfko_kick]
#         )
#
#     line.insert_element(
#         element=rfko_exciter,
#         name='EXCITER1',
#         index='pr.kfb97'
#     )
#     if out_params:
#         params['frev'] = frev
#         return params
#



def insert_rfko_custom_(rfko, signal=None, tune_factor=1 / 3, gain=None, total_signal=True, monofreq=False,
                       freq_start_ratio=None, freq_end_ratio=None, sampling_freq=None,signal_check=True,
                       duration=None, ctx=None):
    ''' TODO 1. check giving directly the signal (with the total length or not), 2 check the monofrequency signal
    '''
    if gain is None:
        gain = rfko.sim_params['gain']
    if freq_end_ratio is None:
        freq_end_ratio= rfko.sim_params['freq_end_ratio']
    if freq_start_ratio is None:
        freq_start_ratio= rfko.sim_params['freq_start_ratio']
    if duration is None:
        duration = rfko.sim_params['duration']
    if sampling_freq is None:
        sampling_freq = rfko.sim_params['sampling_freq']
    if ctx is None:
        ctx = rfko.sim_params['ctx']

    line = rfko.line

    pc = rfko.line_params['pc']
    beta = rfko.line_params['beta']
    Brho = rfko.line_params['Brho']
    charge = rfko.line_params['charge']
    rfko_kick = slwex.kick_angle(float(gain), Brho, pc, beta,charge)

    try:
        twiss = rfko.twiss
    except:
        twiss = rfko.line.twiss(method='4d')

    # revolution freq
    frev = 1/twiss.T_rev0
    time = rfko.sim_params['time']

    start_freq = frev * tune_factor * freq_start_ratio
    end_freq = frev * tune_factor * freq_end_ratio

    line.unfreeze()

    if signal is not None:
        plt.title('custom signal fed to exciter')
        plt.plot(np.arange(len(signal)) / sampling_freq, signal)

        if total_signal:
            # Signal already provided with a length matching the total simulation
            rfko_exciter = xt.Exciter(
                _context=ctx,
                samples=signal,
                sampling_frequency=sampling_freq,
                frev=frev,
                start_turn=0,
                knl=[rfko_kick]
            )
        else:
            rfko_exciter = xt.Exciter(
                _context=ctx,
                samples=signal,
                sampling_frequency=sampling_freq,
                frev=frev,
                duration=float(time),
                start_turn=0,
                knl=[rfko_kick]
            )

    if signal is None:
        if monofreq:

            # To check
            t = np.linspace(0,time,int(sampling_freq*time))
            signal = np.cos(2*np.pi*frev*tune_factor*t)
            if signal_check:
                _,_,f = ph.Ft(signal,t[1],peaks=1,ret_freqs=True,plot=False)
                rfko.logger.warning(f'This is a check of the frequency on the rf signal,'
                                    f'desired f = {frev*tune_factor} \n FFT of the signal {f}')
        else:
            t, signal = ph.generate_chirp(start_freq, end_freq, duration, sampling_freq)

        # Exciter config; the duration parameter tells it to repeat the chirp_signal
        rfko_exciter = xt.Exciter(
            _context=ctx,
            samples=signal,
            sampling_frequency=sampling_freq,
            frev=frev,
            duration=float(time),
            start_turn=0,
            knl=[rfko_kick]
        )



    line.insert_element(
        element=rfko_exciter,
        name='EXCITER1',
        index='pr.kfb97'
    )

    return None




















def plot_kobayashi(dq, S, traj_num=10, npointsXtraj=1000, plot_lims=False, H_max=2, plot_size_ratio=1
                    ):
    ''' Plot kobayashi's Hamiltonia contours
    H max defines the maximum of the Action we want to plot in UNITS of the Stable amplitude'''

    def implicit_function(x, px, dq=dq, S=S):
        # x and px must be normalized
        return 3 * np.pi * dq * (x ** 2 + px ** 2) + 1 / 4 * S * (3 * x * px ** 2 - x ** 3)

    epsilon = 6 * np.pi * dq
    Hsep = ((4 * np.pi * dq) ** 3) / (S ** 2)
    # H_stable = np.array([0,Hsep])
    Hs_in = np.linspace(0, Hsep, traj_num // 2 + 1)
    Hs_out = np.linspace(Hsep, H_max * Hsep, traj_num // 2 + 1)

    X_lim = np.sort([-2 * epsilon / (3 * S), (4 * epsilon / (3 * S))])
    PX_lim = np.sort([-2 * epsilon / (np.sqrt(3) * S), 2 * epsilon / (np.sqrt(3) * S)])

    ## For the plot
    # Define the gradual color palette
    colors1 = [(0, 'green'), (0.5, 'cyan'), (1, 'blue')]  # Example color transitions
    cmap1 = LinearSegmentedColormap.from_list('gradual_palette', colors1)
    colors2 = [(0, 'orange'), (1, 'red')]  # Example color transitions
    cmap2 = LinearSegmentedColormap.from_list('gradual_palette', colors2)

    X_ = np.linspace(-X_lim[1] * 2 * (1 + plot_size_ratio), X_lim[1] * 2 * (1 + plot_size_ratio), npointsXtraj + 1)
    PX_ = np.linspace(-(1 + plot_size_ratio) * PX_lim[1], PX_lim[1] * (1 + plot_size_ratio), npointsXtraj + 1)
    X, Y = np.meshgrid(X_, PX_)
    # Evaluate the implicit function
    Z = implicit_function(X, Y)
    plt.contour(X, Y, Z, levels=np.sort(Hs_in), cmap=cmap1, linewidths=0.5)  # Contour at Z = 0
    ######## external

    X_ = np.linspace(-X_lim[1] * 2 * (1 + plot_size_ratio), X_lim[1] * 2 * (1 + plot_size_ratio), npointsXtraj + 1)
    PX_ = np.linspace(-(1 + plot_size_ratio) * PX_lim[1], PX_lim[1] * (1 + plot_size_ratio), npointsXtraj + 1)
    X, Y = np.meshgrid(X_, PX_)
    # Evaluate the implicit function
    Z = implicit_function(X, Y)
    plt.contour(X, Y, Z, levels=np.sort(Hs_out), cmap=cmap2, linewidths=0.5)  # Contour at Z = 0

    if plot_lims:
        plt.axvline(X_lim[0])
        plt.axvline(X_lim[1])
        plt.axhline(PX_lim[0])
        plt.axhline(PX_lim[1])
    plt.xlabel('X')
    plt.ylabel('Y')
    plt.title('Contour Plot of Implicit Function')
    plt.grid(True)
    plt.axis('equal')  # Equal aspect ratio
    plt.show()


def plot_kobayashi_simple(dq, S, traj_num=10, npointsXtraj=1000, plot_lims=False, H_max=2, plot_size_ratio=1,
                   outside_stable=True):
    ''' H max defines the maximum of the Action we want to plot in UNITS of the Stable amplitude'''

    def implicit_function(x, px, dq=dq, S=S, H=-33e-8):
        # x and px must be normalized
        return 3 * np.pi * dq * (x ** 2 + px ** 2) + 1 / 4 * S * (3 * x * px ** 2 - x ** 3) - H

    epsilon = 6 * np.pi * dq
    Hsep = ((4 * np.pi * dq) ** 3) / (S ** 2)
    # H_stable = np.array([0,Hsep])
    Hs_in = np.linspace(0, Hsep, traj_num // 2 + 1)
    Hs_out = np.linspace(Hsep, H_max * Hsep, traj_num // 2 + 1)
    Xs = []
    PXs = []
    X_lim = np.sort([-2 * epsilon / (3 * S), (4 * epsilon / (3 * S))])
    PX_lim = np.sort([-2 * epsilon / (np.sqrt(3) * S), 2 * epsilon / (np.sqrt(3) * S)])

    ## For the plot
    # Define the gradual color palette
    colors = [(0, 'blue'), (0.5, 'green'), (1, 'red')]  # Example color transitions
    cmap = LinearSegmentedColormap.from_list('gradual_palette', colors)

    for i, H in enumerate(Hs_in):
        # For the version of stable amplitudes inside the stable triangle
        if outside_stable:
            X_ = np.linspace(-X_lim[1] * 2 * (1 + plot_size_ratio), X_lim[1] * 2 * (1 + plot_size_ratio),
                             npointsXtraj + 1)
            PX_ = np.linspace(-(1 + plot_size_ratio) * PX_lim[1], PX_lim[1] * (1 + plot_size_ratio), npointsXtraj + 1)
        else:
            X_ = np.linspace(X_lim[0], X_lim[1], npointsXtraj + 1)
            PX_ = np.linspace(PX_lim[0], PX_lim[1], npointsXtraj + 1)
        X, Y = np.meshgrid(X_, PX_)
        # Evaluate the implicit function
        Z = implicit_function(X, Y, H=H)
        plt.contour(X, Y, Z, levels=[0], colors='r', linewidths=0.5)  # Contour at Z = 0

    for i, H in enumerate(Hs_out):
        X_ = np.linspace(-X_lim[1] * 2 * (1 + plot_size_ratio), X_lim[1] * 2 * (1 + plot_size_ratio), npointsXtraj + 1)
        PX_ = np.linspace(-(1 + plot_size_ratio) * PX_lim[1], PX_lim[1] * (1 + plot_size_ratio), npointsXtraj + 1)
        X, Y = np.meshgrid(X_, PX_)
        # Evaluate the implicit function
        Z = implicit_function(X, Y, H=H)
        plt.contour(X, Y, Z, levels=[0], colors='b', linewidths=0.5)  # Contour at Z = 0
    if plot_lims:
        plt.axvline(X_lim[0])
        plt.axvline(X_lim[1])
        plt.axhline(PX_lim[0])
        plt.axhline(PX_lim[1])
    plt.xlabel('X')
    plt.ylabel('Y')
    plt.title('Contour Plot of Implicit Function')
    plt.grid(True)
    plt.axis('equal')  # Equal aspect ratio
    plt.show()


def kobay_vs_tracking_1(rfko, tune_corr=False, pdigit=3):
    ''' to start simple it will consider a particle. Package formalism

   '''
    assert hasattr(rfko, 'monitor'), 'No tracked particle'

    if not hasattr(rfko, 'twiss'):
        rfko.twiss = rfko.line.twiss()

    ## Temporary for 1 particle
    assert rfko.sim_params['n_part'] == 1, 'This procedure is reserved for 1 particle at the moment'

    element = rfko.monitor.placed_at_element
    outdict = rfko.to_dict()
    Monitor = xh.Normalize_monitor(outdict)

    #### I ROTATE THE TRACKED PARTICLES to bring them back at the correct orientation of the Hamiltonian
    tri_g = slwex.tri_geom(outdict, negate_s=True, sqrt=tune_corr, interpolate=True)
    tune_dist = tri_g['tune_distance']
    nominal_dq = rfko.twiss.qx - 6 - 1 / 3
    S = tri_g['S']
    dmu = rfko.twiss.mux[rfko.twiss.name == element] * 2 * np.pi - tri_g['S_mu']

    # Plot pos 1
    plt.figure(figsize=(10, 8))
    plt.scatter(Monitor.x, Monitor.px)
    plt.title('Normalized')

    # Rotate coordinate of the tracked particle
    X, PX = ph.rotate(Monitor.x[0], Monitor.px[0], -dmu[0],
                   center_mean=True)  # the mean_center shouldn't change the situation
    ### Plot pos2 VS
    plt.figure(figsize=(10, 8))
    plt.scatter(X, PX)
    plt.title(f"Debug ph.rotated of {dmu} (pi)")
    ######### calculate Hs
    Hs = np.mean(slwex.kobay(X, PX, dq=tune_dist, S=S))
    print(Hs)
    order_Hs = np.log10(np.abs(Hs))
    Hs_cut = round(float(Hs), int(np.abs(order_Hs)) + pdigit)

    #### PLot H
    plt.figure()
    plt.plot(slwex.kobay(X, PX) / Hs)
    plt.title('H-values')

    ### PLOT THE CORRESPONDING CONTOURPLOT AND THEN ON TOP, THE TRACKED PARTICLES
    margin_ = 1.2  # margin of the contour plot
    X_ = np.linspace(np.min(X) * margin_, np.max(X) * margin_, 2000)
    PX_ = np.linspace(margin_ * np.min(PX), margin_ * np.max(PX), 2000)
    meshx, meshpx = np.meshgrid(X_, PX_)
    # print(Hs)

    Hmesh = slwex.kobay(meshx, meshpx, dq=tune_dist, S=S)

    # BAKUP POINT
    plt.figure()
    plt.contour(meshx, meshpx, Hmesh, levels=[Hs])
    plt.scatter(X, PX, s=2)
    print(f'The nominal tune distance is {nominal_dq} \n The one with the correction to the orbit is {tune_dist}')
    plt.show()


def kobay_vs_tracking_n(rfko, tune_sqrt=False):
    ''' to start simple it will consider a particle. Package formalism

   '''
    assert hasattr(rfko, 'monitor'), 'No tracked particle'
    assert rfko.sim_params['DPP_FACTOR'] == 0, 'The corrections for the off momentum particles is wrong still'
    if (not hasattr(rfko, 'twiss'))&(rfko.line_type!='henon_map'):
        rfko.twiss = rfko.line.twiss()

    element = rfko.monitor.placed_at_element
    outdict = rfko.to_dict()
    Monitor = xh.Normalize_monitor(outdict)

    #### I ROTATE THE TRACKED PARTICLES to bring them back at the correct orientation of the Hamiltonian
    if rfko.line_type != 'henon_map':
        tri_g = slwex.tri_geom(outdict, negate_s=True, sqrt=tune_sqrt, interpolate=True)
        tune_dist = tri_g['tune_distance']
        nominal_dq = rfko.twiss.qx - 6 - 1 / 3
        S = tri_g['S']
        dmu = rfko.twiss.mux[rfko.twiss.name == element] * 2 * np.pi - tri_g['S_mu']
    else:
        dmu = 0 # monitor is at the virtual sextupole
        tune_dist = rfko.henon_params['qx']-6-1/3
        nominal_dq = tune_dist
        # Without the dispersion the nomila tune distance is the same as the tune distance
        S = rfko.henon_params['S']

    # Rotate coordinate of the tracked particle
    X, PX = ph.rotate(Monitor.x, Monitor.px, -dmu[0],
                   center_mean=True)
    # Rotation sign is for diplacement from the sextupole so to go back sign need to be inverted
    # the mean_center shouldn't change the situation since the traj are already centered

    # Mean value of the hamiltonian (whole trajectorie) for every particle
    Hs = np.mean(slwex.kobay(X, PX, dq=tune_dist, S=S), axis=1)

    ### PLOT THE CORRESPONDING CONTOURPLOT AND THEN ON TOP, THE TRACKED PARTICLES
    margin_ = 1.2  # margin of the contour plot
    X_ = np.linspace(np.min(X) * margin_, np.max(X) * margin_, 2000)
    PX_ = np.linspace(margin_ * np.min(PX), margin_ * np.max(PX), 2000)
    meshx, meshpx = np.meshgrid(X_, PX_)

    Hmesh = slwex.kobay(meshx, meshpx, dq=tune_dist, S=S)
    colors = [(0, 'blue'), (0.5, 'green'), (1, 'red')]  # Example color transitions
    cmap = LinearSegmentedColormap.from_list('gradual_palette', colors)
    H_color = (Hs - np.min(Hs)) / np.abs(np.max(Hs - np.min(Hs)))
    print(Hs)
    print(H_color)
    # BAKUP POINT
    plt.figure(figsize=(10, 8))
    plt.contour(meshx, meshpx, Hmesh, levels=np.sort(Hs), cmap=cmap)
    colors = cmap(H_color)
    for i in range(X.shape[0]):
        plt.scatter(X[i], PX[i], s=2, color=colors[i])
    print(f'The nominal tune distance is {nominal_dq} \n The one with the correction to the orbit is {tune_dist}')
    plt.show()